"""
Contains general calculation and plotting tools.
"""
import json
import logging
import math
import os
import random
from bisect import bisect
from concurrent.futures import ProcessPoolExecutor, as_completed
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from dataclasses import dataclass
from enum import Enum, unique
from functools import wraps
from io import StringIO
from itertools import accumulate, chain, compress, zip_longest
from pathlib import Path
from typing import Any, Callable, Dict, List, Sequence, Set, Tuple, Union, overload

import geopandas as gpd
import numpy as np
import pandas as pd
import pygeos
import sklearn.metrics as sklm
from geopandas.sindex import PyGEOSSTRTreeIndex
from joblib import Memory
from matplotlib import patheffects as path_effects
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from shapely import prepared, wkb
from shapely.affinity import scale
from shapely.geometry import (
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
    box,
)
from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry
from shapely.ops import split
from sklearn.linear_model import LinearRegression

log = logging.getLogger(__name__)

styled_text_dict = {
    "path_effects": [path_effects.withStroke(linewidth=3, foreground="k")],
    "color": "white",
}
styled_prop = dict(
    boxstyle="round",
    pad=0.6,
    facecolor="wheat",
    path_effects=[path_effects.SimplePatchShadow(), path_effects.Normal()],
)

# Columns for report_df

CC_branch = "C - C"
CE_branch = "C - E"
CI_branch = "C - I"
IE_branch = "I - E"
II_branch = "I - I"
EE_branch = "E - E"
Error_branch = "Error"
X_node = "X"
Y_node = "Y"
I_node = "I"
E_node = "E"
BOUNDARY_INTERSECT_KEYS = ("0", "1", "2")

CONNECTION_COLUMN = "Connection"
CLASS_COLUMN = "Class"
GEOMETRY_COLUMN = "geometry"
POWERLAW = "powerlaw"
LOGNORMAL = "lognormal"
EXPONENTIAL = "exponential"

NULL_SET = "-1"

REPRESENTATIVE_POINT = "Representative Point"
RADIUS = "Radius"
NAME = "Name"
CENSORING = "Censoring"
RELATIVE_CENSORING = "Relative Censoring"

GEOJSON_DRIVER = "GeoJSON"

SetRangeTuple = Tuple[Tuple[float, float], ...]
BoundsTuple = Tuple[float, float, float, float]
PointTuple = Tuple[float, float]
Number = Union[float, int]
ParameterValuesType = Dict[str, Union[float, int, str]]
ParameterListType = List[ParameterValuesType]

MINIMUM_LINE_LENGTH = 1e-18

DEFAULT_FRACTOPO_CACHE_PATH = Path(".cache/fractopo")

JOBLIB_CACHE = Memory(
    os.environ.get("FRACTOPO_CACHE_PATH", DEFAULT_FRACTOPO_CACHE_PATH)
    # Disk caching is disabled in pytest execution by setting an environment
    # variable FRACTOPO_DISABLE_CACHE to a non-zero string (i.e. "1") However,
    # it can be enabled even in tests by setting FRACTOPO_DISABLE_CACHE to 0
    if os.environ.get("FRACTOPO_DISABLE_CACHE") in (None, "0") else None,
    verbose=int(os.environ.get("FRACTOPO_JOBLIB_CACHE_VERBOSITY", 0)),
)


@dataclass
class ProcessResult:

    """
    Dataclass for multiprocessing result parsing.
    """

    identifier: str
    result: Any
    error: bool


@unique
class Col(Enum):

    """
    GeoDataFrame column names for attributes.
    """

    LENGTH = "length"
    AZIMUTH = "azimuth"
    AZIMUTH_SET = "azimuth_set"
    LENGTH_SET = "length_set"
    LENGTH_WEIGHTS = "boundary_weight"
    LENGTH_NON_WEIGHTED = "length non-weighted"


def sum_aggregation(values, **_) -> Number:
    """
    Aggregate by calculating sum.
    """
    value_sum = np.array(values).sum()
    assert isinstance(value_sum, (float, int))
    return value_sum


def mean_aggregation(values, weights) -> Number:
    """
    Aggregate by calculating mean.
    """
    average_value = np.average(values, weights=weights)
    assert isinstance(average_value, (float, int))
    return average_value


def fallback_aggregation(values) -> str:
    """
    Fallback aggregation where values are simply joined into a string.
    """
    return str(values)


@unique
class Aggregator(Enum):

    """
    Define how to aggregate during subsample aggragation.
    """

    SUM = sum_aggregation
    MEAN = mean_aggregation


@dataclass
class ParamInfo:

    """
    Parameter with name and metadata.
    """

    name: str
    plot_as_log: bool
    unit: str
    # TODO: Currently not used.
    needs_topology: bool
    aggregator: Aggregator


@unique
class Param(Enum):

    """
    Column names for geometric and topological parameters.

    The ``ParamInfo`` instances contain additional metadata.
    """

    AREA = ParamInfo("Area", False, r"$m^2$", False, Aggregator.SUM)
    AREAL_FREQUENCY_B20 = ParamInfo(
        "Areal Frequency B20", True, r"$\frac{1}{m^2}$", True, Aggregator.MEAN
    )
    AREAL_FREQUENCY_P20 = ParamInfo(
        "Areal Frequency P20", True, r"$\frac{1}{m^2}$", True, Aggregator.MEAN
    )
    BRANCH_MEAN_LENGTH = ParamInfo(
        "Branch Mean Length", True, "$m$", True, Aggregator.MEAN
    )
    BRANCH_MIN_LENGTH = ParamInfo(
        "Branch Min Length", True, "$m$", True, Aggregator.MEAN
    )
    BRANCH_MAX_LENGTH = ParamInfo(
        "Branch Max Length", True, "$m$", True, Aggregator.MEAN
    )
    CONNECTIONS_PER_BRANCH = ParamInfo(
        "Connections per Branch", False, r"$\frac{1}{n}$", True, Aggregator.MEAN
    )
    CONNECTIONS_PER_TRACE = ParamInfo(
        "Connections per Trace", False, r"$\frac{1}{n}$", True, Aggregator.MEAN
    )
    CONNECTION_FREQUENCY = ParamInfo(
        "Connection Frequency", False, r"$\frac{1}{m^2}$", True, Aggregator.MEAN
    )
    DIMENSIONLESS_INTENSITY_B22 = ParamInfo(
        "Dimensionless Intensity B22", False, "-", True, Aggregator.MEAN
    )
    DIMENSIONLESS_INTENSITY_P22 = ParamInfo(
        "Dimensionless Intensity P22", False, "-", False, Aggregator.MEAN
    )
    FRACTURE_DENSITY_MAULDON = ParamInfo(
        "Fracture Density (Mauldon)", True, r"$\frac{1}{m^2}$", True, Aggregator.MEAN
    )
    FRACTURE_INTENSITY_B21 = ParamInfo(
        "Fracture Intensity B21", True, r"$\frac{m}{m^2}$", False, Aggregator.MEAN
    )
    FRACTURE_INTENSITY_MAULDON = ParamInfo(
        "Fracture Intensity (Mauldon)", True, r"$\frac{m}{m^2}$", True, Aggregator.MEAN
    )
    FRACTURE_INTENSITY_P21 = ParamInfo(
        "Fracture Intensity P21", True, r"$\frac{m}{m^2}$", False, Aggregator.MEAN
    )
    NUMBER_OF_BRANCHES = ParamInfo(
        "Number of Branches", False, "-", True, Aggregator.SUM
    )
    NUMBER_OF_BRANCHES_TRUE = ParamInfo(
        "Number of Branches (Real)", False, "-", True, Aggregator.SUM
    )
    NUMBER_OF_TRACES = ParamInfo("Number of Traces", False, "-", True, Aggregator.SUM)
    NUMBER_OF_TRACES_TRUE = ParamInfo(
        "Number of Traces (Real)", False, "-", True, Aggregator.SUM
    )
    TRACE_MEAN_LENGTH = ParamInfo(
        "Trace Mean Length", True, "$m$", False, Aggregator.MEAN
    )
    TRACE_MIN_LENGTH = ParamInfo(
        "Trace Min Length", True, "$m$", False, Aggregator.MEAN
    )
    TRACE_MAX_LENGTH = ParamInfo(
        "Trace Max Length", True, "$m$", False, Aggregator.MEAN
    )
    TRACE_MEAN_LENGTH_MAULDON = ParamInfo(
        "Trace Mean Length (Mauldon)", True, "$m$", True, Aggregator.MEAN
    )
    CIRCLE_COUNT = ParamInfo(
        "Circle Count",
        plot_as_log=False,
        unit="-",
        needs_topology=False,
        aggregator=Aggregator.SUM,
    )


def determine_set(
    value: float,
    value_ranges: SetRangeTuple,
    set_names: Tuple[str, ...],
    loop_around: bool,
) -> str:
    """
    Determine which named value range, if any, value is within.

    loop_around defines behavior expected for radial data i.e. when value range
    can loop back around e.g. [160, 50]

    E.g.

    >>> determine_set(10.0, [(0, 20), (30, 160)], ["0-20", "30-160"], False)
    '0-20'

    Example with

    >>> determine_set(50.0, [(0, 20), (160, 60)], ["0-20", "160-60"], True)
    '160-60'

    :param value: Value to determine set of.
    :param value_ranges: Ranges of each set.
    :param set_names: Names of each set.
    :param loop_around: Whether the sets loop around. This is the case for
        radial data such as azimuths but not the case for length data.
    :return: Set string in which value belongs.
    :raises ValueError: When set value ranges overlap.
    """
    assert len(value_ranges) == len(set_names)
    possible_set_name = [
        set_name
        for set_name, value_range in zip(set_names, value_ranges)
        if is_set(value, value_range, loop_around)
    ]
    if len(possible_set_name) == 0:
        return NULL_SET
    if len(possible_set_name) == 1:
        return possible_set_name[0]
    raise ValueError("Expected set value ranges to not overlap.")


def is_set(value: Number, value_range: Tuple[float, float], loop_around: bool) -> bool:
    """
    Determine if value fits within the given value_range.

    If the value range has the possibility of looping around loop_around can be
    set to true.

    >>> is_set(5, (0, 10), False)
    True

    >>> is_set(5, (175, 15), True)
    True

    :param value: Value to determine.
    :param value_range Tuple[float,: The range of values.
    :param loop_around: Whether the range loops around. This is the case for
        radial data such as azimuths but not the case for length data.
    :return: Is it within range.
    """
    if loop_around:
        if value_range[0] > value_range[1]:
            # Loops around case
            if (value >= value_range[0]) | (value <= value_range[1]):
                return True
    if value_range[0] <= value <= value_range[1]:
        return True
    return False


def is_azimuth_close(
    first: float, second: float, tolerance: float, halved: bool = True
):
    """
    Determine are azimuths first and second within tolerance.

    Takes into account the radial nature of azimuths.

    >>> is_azimuth_close(0, 179, 15)
    True

    >>> is_azimuth_close(166, 179, 15)
    True

    >>> is_azimuth_close(20, 179, 15)
    False

    :param first: First azimuth value to compare.
    :param second: Second azimuth value to compare.
    :param tolerance: Tolerance for closeness.
    :param halved: Are the azimuths azial (i.e. ``halved=True``) or vectors.
    """
    diff = abs(first - second)
    if halved:
        diff = diff if diff <= 90 else 180 - diff
        assert 0 <= diff <= 90
    else:
        diff = diff if diff <= 180 else 360 - diff
        assert 0 <= diff <= 180
    return diff < tolerance


def determine_regression_azimuth(line: LineString) -> float:
    """
    Determine azimuth of line LineString with linear regression.

    A scikit-learn LinearRegression is fitted to the x, y coordinates of the
    given  and the azimuth of the fitted linear line is returned.

    The azimuth is returned in range [0, 180].

    E.g.

    >>> line = LineString([(0, 0), (1, 1), (2, 2), (3, 3)])
    >>> determine_regression_azimuth(line)
    45.0

    >>> line = LineString([(-1, -5), (3, 3)])
    >>> round(determine_regression_azimuth(line), 3)
    26.565

    >>> line = LineString([(0, 0), (0, 3)])
    >>> determine_regression_azimuth(line)
    0.0


    :param line: The line of which azimuth is determined.
    :return: The determined azimuth in range [0, 180].
    :raises ValueError: When ``LinearRegression`` returns
        unexpected coefficients.
    """
    x, y = zip(*line.coords)
    x = np.array(x).reshape((-1, 1))
    y = np.array(y)
    model = LinearRegression().fit(x, y)
    coefs = model.coef_
    if len(coefs) == 1:
        coef = coefs[0]
    else:
        raise ValueError("Expected model.coef_ to be an array of length 1.")
    if np.isclose(coef, 0.0):
        return 0.0
    azimuth = np.rad2deg(np.arctan(1 / coef))  # type: ignore
    if azimuth < 0:
        azimuth = 90 - np.rad2deg(np.arctan(coef))  # type: ignore
    assert isinstance(azimuth, float) and 0 <= azimuth <= 180
    return azimuth


def determine_azimuth(line: LineString, halved: bool) -> float:
    """
    Calculate azimuth of given line.

    If halved -> return is in range [0, 180]
    Else -> [0, 360]

    e.g.:
    Accepts LineString

    >>> determine_azimuth(LineString([(0, 0), (1, 1)]), True)
    45.0

    >>> determine_azimuth(LineString([(0, 0), (0, 1)]), True)
    0.0

    >>> determine_azimuth(LineString([(0, 0), (-1, -1)]), False)
    225.0

    >>> determine_azimuth(LineString([(0, 0), (-1, -1)]), True)
    45.0

    :param line: The line of which azimuth is determined.
    :param halved: Whether to return result in range [0, 180]
        (halved=True) or [0, 360] (halved=False).
    :return: The determined azimuth.
    """
    coord_list = list(line.coords)
    start_x = coord_list[0][0]
    start_y = coord_list[0][1]
    end_x = coord_list[-1][0]
    end_y = coord_list[-1][1]
    azimuth = 90 - math.degrees(math.atan2((end_y - start_y), (end_x - start_x)))
    if azimuth < 0:
        azimuth = azimuth + 360
    if azimuth > 360:
        azimuth -= 360
    if halved:
        azimuth = azimuth if not azimuth >= 180 else azimuth - 180
    return azimuth


def calc_strike(dip_direction: float) -> float:
    """
    Calculate strike from dip direction. Right-handed rule.

    E.g.:

    >>> calc_strike(50.0)
    320.0

    >>> calc_strike(180.0)
    90.0

    :param dip_direction: The direction of dip.
    :return: Converted strike.
    """
    strike = dip_direction - 90
    if strike < 0:
        strike = 360 + strike
    return strike


def azimu_half(degrees: float) -> float:
    """
    Transform azimuth from 180-360 range to range 0-180.

    :param degrees: Degrees in range 0 - 360
    :return: Degrees in range 0 - 180
    """
    if degrees >= 180:
        degrees = degrees - 180
    return degrees


def sd_calc(data):
    """
    Calculate standard deviation for radial data.

    TODO: Wrong results atm. Needs to take into account real length, not just
    orientation of unit vector.  Calculates standard deviation for radial data
    (degrees)

    E.g.

    >>> sd_calc(np.array([2, 5, 8]))
    (3.0, 5.00)

    :param data: Array of degrees
    :type data: np.ndarray
    :return: Standard deviation
    """
    n = len(data)
    if n <= 1:
        return 0.0
    mean, sd = avg_calc(data), 0.0
    # calculate stan. dev.
    for el in data:
        diff = abs(mean - float(el))
        if diff > 180:
            diff = 360 - diff

        sd += diff**2
    sd = math.sqrt(sd / float(n - 1))

    return sd, mean


def avg_calc(data):
    """
    Calculate average for radial data.

    TODO: Should take length into calculations.......................... not
    real average atm
    """
    n, mean = len(data), 0.0

    if n <= 1:
        return data[0]
    vectors = []
    # calculate average
    for el in data:
        rad = math.radians(el)
        v = np.array([math.cos(rad), math.sin(rad)])
        vectors.append(v)

    meanv = np.mean(np.array(vectors), axis=0)

    mean = math.degrees(math.atan2(meanv[1], meanv[0]))
    # print(mean)
    if mean < 0:
        mean = 360 + mean
    return mean


def define_length_set(length: float, set_df: pd.DataFrame) -> str:
    """
    Define sets based on the length of the traces or branches.
    """
    if length < 0:
        raise ValueError("Length value was not positive.\n Value: {length}")

    # Set_num is -1 when length is in no length set range
    set_label = -1
    for _, row in set_df.iterrows():
        set_range: Tuple[float, float] = row.LengthSetLimits
        # Checks if length degree value is within the length set range
        if set_range[0] <= length <= set_range[1]:
            set_label = row.LengthSet

    return str(set_label)


# def curviness(linestring):
#     """
#     Determine curviness of LineString.
#     """
#     try:
#         coords = list(linestring.coords)
#     except NotImplementedError:
#         return np.NaN
#     df = pd.DataFrame(columns=["azimu", "length"])
#     for i in range(len(coords) - 1):
#         start = Point(coords[i])
#         end = Point(coords[i + 1])
#         line = LineString([start, end])
#         azimu = determine_azimuth(line, halved=True)
#         # halved = tools.azimu_half(azimu)
#         length = line.length
#         addition = {
#             "azimu": azimu,
#             "length": length,
#         }  # Addition to DataFrame with fit x and fit y values
#         df = df.append(addition, ignore_index=True)

#     std = sd_calc(df.azimu.values.tolist())
#     azimu_std = std

#     return azimu_std


def prepare_geometry_traces(trace_series: gpd.GeoSeries) -> prepared.PreparedGeometry:
    """
    Prepare trace_series geometries for a faster spatial analysis.

    Assumes geometries are LineStrings which are consequently collected into a
    single MultiLineString which is prepared with shapely.prepared.prep.

    >>> traces = gpd.GeoSeries(
    ...     [LineString([(0, 0), (1, 1)]), LineString([(0, 1), (0, -1)])]
    ... )
    >>> prepare_geometry_traces(traces).context.wkt
    'MULTILINESTRING ((0 0, 1 1), (0 1, 0 -1))'
    """
    traces = trace_series.geometry.values
    traces = np.asarray(traces).tolist()  # type: ignore
    trace_col = MultiLineString(traces)
    prepared_traces = prepared.prep(trace_col)
    return prepared_traces


def match_crs(
    first: Union[gpd.GeoSeries, gpd.GeoDataFrame],
    second: Union[gpd.GeoSeries, gpd.GeoDataFrame],
) -> Tuple[
    Union[gpd.GeoSeries, gpd.GeoDataFrame], Union[gpd.GeoSeries, gpd.GeoDataFrame]
]:
    """
    Match crs between two geopandas data structures.

    >>> first = gpd.GeoSeries([Point(1, 1)], crs="EPSG:3067")
    >>> second = gpd.GeoSeries([Point(1, 1)])
    >>> m_first, m_second = match_crs(first, second)
    >>> m_first.crs == m_second.crs
    True
    """
    all_crs = [crs for crs in (first.crs, second.crs) if crs is not None]
    if len(all_crs) == 0:
        # No crs in either input
        return first, second
    if len(all_crs) == 2 and all_crs[0] == all_crs[1]:
        # Two valid and they're the same
        return first, second
    if len(all_crs) == 1:
        # One valid crs in inputs
        crs = all_crs[0]
        first.crs = crs
        second.crs = crs
        return first, second
    # Two crs that are not the same
    return first, second


def replace_coord_in_trace(
    trace: LineString, index: int, replacement: Point
) -> LineString:
    """
    Replace coordinate Point of LineString at index with replacement Point.
    """
    coord_points = get_trace_coord_points(trace)
    coord_points.pop(index)
    coord_points.insert(index, replacement)
    new_linestring = LineString(coord_points)
    return new_linestring


def get_next_point_in_trace(trace: LineString, point: Point) -> Point:
    """
    Determine next coordinate point towards middle of LineString from point.
    """
    coord_points = get_trace_coord_points(trace)
    assert point in coord_points
    if point == coord_points[-1]:
        return coord_points[-2]
    if point == coord_points[0]:
        return coord_points[1]
    raise ValueError("Expected point to be a coord in trace.")


def get_trace_endpoints(
    trace: LineString,
) -> Tuple[Point, Point]:
    """
    Return endpoints (shapely.geometry.Point) of a given LineString.
    """
    if not isinstance(trace, LineString):
        raise TypeError(
            "Non LineString geometry passed into get_trace_endpoints.\n"
            f"trace: {trace}"
        )
    points = Point(trace.coords[0]), Point(trace.coords[-1])
    return points


def get_trace_coord_points(trace: LineString) -> List[Point]:
    """
    Get all coordinate Points of a LineString.

    >>> trace = LineString([(0, 0), (2, 0), (3, 0)])
    >>> coord_points = get_trace_coord_points(trace)
    >>> print([p.wkt for p in coord_points])
    ['POINT (0 0)', 'POINT (2 0)', 'POINT (3 0)']

    """
    assert isinstance(trace, LineString)
    return [Point(xy) for xy in trace.coords]


def point_to_xy(point: Point) -> Tuple[float, float]:
    """
    Get x and y coordinates of Point.
    """
    x, y = point.xy
    x, y = [val[0] for val in (x, y)]
    return (x, y)


@JOBLIB_CACHE.cache
def determine_general_nodes(
    traces: Union[gpd.GeoSeries, gpd.GeoDataFrame]
) -> Tuple[List[Tuple[Point, ...]], List[Tuple[Point, ...]]]:
    """
    Determine points of intersection and endpoints of given trace LineStrings.

    The points are linked to the indexes of the traces.
    """
    if traces.shape[0] == 0 or traces.empty:
        raise ValueError("Expected non-empty dataset in determine_nodes.")
    if not (
        traces.index.values[0] == 0 and traces.index.values[-1] == traces.shape[0] - 1
    ):
        raise ValueError(
            "Expected traces to have a continuous index from 0...len(traces) - 1"
        )

    intersect_nodes: List[Tuple[Point, ...]] = []
    endpoint_nodes: List[Tuple[Point, ...]] = []
    # spatial_index = traces.geometry.sindex
    try:
        spatial_index = pygeos_spatial_index(traces.geometry)
    except TypeError:
        spatial_index = None
    for idx, geom in enumerate(traces.geometry.values):
        if not isinstance(geom, LineString) or geom.is_empty:
            # Intersections and endpoints are not defined for
            # MultiLineStrings or empty LineStrings
            intersect_nodes.append(())
            endpoint_nodes.append(())
            continue

        # Get trace candidates for intersection
        trace_candidates_idx: List[int] = (
            sorted(spatial_index_intersection(spatial_index, geom_bounds(geom)))
            if spatial_index is not None
            else [idx]
        )

        # Remove current geometry from candidates
        trace_candidates_idx.remove(idx)
        trace_candidates: gpd.GeoSeries = traces.geometry.iloc[trace_candidates_idx]

        # Only accept LineString candidates
        trace_candidates = trace_candidates.loc[
            [isinstance(geom, LineString) for geom in trace_candidates.geometry.values]
        ]

        intersection_geoms = determine_valid_intersection_points_no_vnode(
            trace_candidates, geom
        )
        intersect_nodes.append(tuple(intersection_geoms))
        endpoints = tuple(
            (
                endpoint
                for endpoint in get_trace_endpoints(geom)
                if not any(
                    # Intersection results in inaccurate geoms ->
                    # tolerance is actually close to a snap_threshold
                    # This represents a hard limit to snapping threshold.
                    np.isclose(endpoint.distance(intersection_geom), 0, atol=1e-3)
                    for intersection_geom in intersection_geoms
                    if not intersection_geom.is_empty
                )
            )
        )
        endpoint_nodes.append(endpoints)
    return intersect_nodes, endpoint_nodes


def determine_valid_intersection_points_no_vnode(
    trace_candidates: gpd.GeoSeries, geom: LineString
) -> List[Point]:
    """
    Filter intersection points between trace candidates and geom with no vnodes.

    V-node intersections are validated by looking at the endpoints. If V-nodes
    were kept as intersection points the VNodeValidator could not find V-node
    errors.
    """
    # Determine intersections between candidate LineStrings and the geom LineString
    inter = determine_valid_intersection_points(trace_candidates.intersection(geom))

    # If empty, exit
    if len(inter) == 0:
        return inter

    # Get endpoints of the geom LineString
    geom_endpoints = get_trace_endpoints(geom)

    # Upkeep a boolean list of which indexes (i.e., valid intersection points)
    # to keep
    # This will be used to compress `inter` before returning
    p_to_keep = [True] * len(inter)

    # Iterate over the trace candidates
    for trace_candidate in trace_candidates.geometry.values:
        candidate_endpoints = get_trace_endpoints(trace_candidate)
        for ce in candidate_endpoints:
            for ge in geom_endpoints:
                # First check: is the candidate endpoint close to the geom endpoint
                candidate_endpoint_is_close_to_geom_endpoint = np.isclose(
                    ce.distance(ge), 0, atol=1e-4
                )
                # If not, keep it in `inter`
                if not candidate_endpoint_is_close_to_geom_endpoint:
                    continue

                # Enumerate over the valid intersection points
                for idx, p in enumerate(inter):
                    # If the current valid intersection point is already set
                    # for removal, continue the loop to the next
                    if not p_to_keep[idx]:
                        continue
                    # Second check: is the `geom` LineString endpoint close to
                    # the intersection point `p`. If yes, mark the intersection
                    # point for removal from `interl`
                    if np.isclose(ge.distance(p), 0, atol=1e-4):
                        p_to_keep[idx] = False
                        # inter.remove(p)
                        # p_to_remove.add(idx)

    # Remove the marked intersection points from `inter` based on the boolean
    # list `p_to_keep`
    inter_filtered = list(compress(inter, selectors=p_to_keep))
    return inter_filtered


def determine_valid_intersection_points(
    intersection_geoms: gpd.GeoSeries,
) -> List[Point]:
    """
    Filter intersection points between trace candidates and geom.

    Only allows Point geometries as intersections. LineString intersections
    would be possible if geometries are stacked.

    :param intersection_geoms: GeoSeries of intersection (Point) geometries.
    :return: The valid intersections (Points).
    """
    assert isinstance(intersection_geoms, gpd.GeoSeries)
    valid_interaction_points = []
    for geom in intersection_geoms.geometry.values:
        assert isinstance(geom, (BaseGeometry, BaseMultipartGeometry))
        if isinstance(geom, Point):
            valid_interaction_points.append(geom)
        elif isinstance(geom, MultiPoint):
            valid_interaction_points.extend(list(geom.geoms))
        elif geom.is_empty:
            log.info(
                f"Empty geometry in determine_valid_intersection_points: {geom.wkt}"
            )

        # TODO: Should these clauses error or report the invalid geometries upstream?
        elif isinstance(geom, LineString):
            log.error(f"Expected geom ({geom.wkt}) not to be of type LineString.")
        elif isinstance(geom, MultiLineString):
            log.error(f"Expected geom ({geom.wkt}) not to be of type MultiLineString.")
        else:
            raise TypeError(
                "Expected (Multi)Point or (Multi)LineString geometries"
                " in determine_valid_intersection_points."
            )
    assert all(isinstance(p, Point) for p in valid_interaction_points)
    return valid_interaction_points


def line_intersection_to_points(first: LineString, second: LineString) -> List[Point]:
    """
    Perform shapely intersection between two LineStrings.

    Enforces only Point returns.
    """
    intersection = first.intersection(second)
    collect_points: List[Point] = []
    if isinstance(intersection, LineString) and intersection.is_empty:
        pass
    elif isinstance(intersection, Point):
        collect_points = [intersection]
    elif isinstance(intersection, MultiPoint):
        collect_points = list(intersection.geoms)
    else:
        log.error(f"Expected Point or empty intersection, got: {intersection}")
    return collect_points


def flatten_tuples(
    list_of_tuples: List[Tuple[Any, ...]]
) -> Tuple[List[int], List[Any]]:
    """
    Flatten collection of tuples and return index references.

    Indexes are from original tuple groupings.

    E.g.

    >>> tuples = [(1, 1, 1), (2, 2, 2, 2), (3,)]
    >>> flatten_tuples(tuples)
    ([0, 0, 0, 1, 1, 1, 1, 2], [1, 1, 1, 2, 2, 2, 2, 3])

    """
    accumulated_idxs = list(
        accumulate([len(val_tuple) for _, val_tuple in enumerate(list_of_tuples)])
    )
    flattened_tuples = list(chain(*list_of_tuples))
    flattened_idx_reference = [
        bisect(accumulated_idxs, idx) for idx in range(len(flattened_tuples))
    ]
    return flattened_idx_reference, flattened_tuples


@JOBLIB_CACHE.cache
def determine_node_junctions(
    nodes: List[Tuple[Point, ...]],
    snap_threshold: float,
    snap_threshold_error_multiplier: float,
    error_threshold: int,
) -> Set[int]:
    """
    Determine faulty trace junctions using nodes.

    E.g.

    >>> nodes = [
    ...     (Point(0, 0), Point(1, 1)),
    ...     (Point(1, 1),),
    ...     (Point(5, 5),),
    ...     (Point(0, 0), Point(1, 1)),
    ... ]
    >>> snap_threshold = 0.01
    >>> snap_threshold_error_multiplier = 1.1
    >>> error_threshold = 2
    >>> determine_node_junctions(
    ...     nodes, snap_threshold, snap_threshold_error_multiplier, error_threshold
    ... )
    {0, 1, 3}
    """
    if len(nodes) == 0:
        return set()

    flattened_idx_reference, flattened_node_tuples = flatten_tuples(nodes)

    if len(flattened_node_tuples) == 0:
        return set()

    # Collect nodes into GeoSeries
    flattened_nodes_geoseries = gpd.GeoSeries(flattened_node_tuples)

    # Create spatial index of nodes
    # nodes_geoseries_sindex = flattened_nodes_geoseries.sindex
    nodes_geoseries_sindex = pygeos_spatial_index(flattened_nodes_geoseries)

    # Set collection for indexes with junctions
    indexes_with_junctions: Set[int] = set()

    # Iterate over node tuples i.e. points is a tuple with Points
    # The node tuple indexes represent the trace indexes
    for idx, points in enumerate(nodes):
        associated_point_count = len(points)
        if associated_point_count == 0:
            continue

        # Because node indexes represent traces, we can remove all nodes of the
        # current trace by using the idx.
        other_nodes_geoseries: gpd.GeoSeries = flattened_nodes_geoseries.loc[
            [idx_reference != idx for idx_reference in flattened_idx_reference]
        ]

        # Iterate over the actual Points of the current trace
        for _, point in enumerate(points):
            # Get node candidates from spatial index
            node_candidates_idx = spatial_index_intersection(
                nodes_geoseries_sindex,
                geom_bounds(
                    safe_buffer(
                        point, snap_threshold * snap_threshold_error_multiplier * 10
                    )
                ),
            )

            # Shift returned indexes by associated_point_count to match to
            # correct points
            remaining_idxs = set(other_nodes_geoseries.index.values)

            # Only choose node candidates that are not part of current traces
            # nodes
            node_candidates_idx = [
                val if val <= idx else val - associated_point_count
                for val in node_candidates_idx
                if val in remaining_idxs
            ]
            node_candidates: gpd.GeoSeries = other_nodes_geoseries.iloc[  # type: ignore
                node_candidates_idx
            ]
            # snap_threshold * snap_threshold_error_multiplier represents the
            # threshold for intersection. Actual intersection is not necessary.
            intersection_data = [
                intersecting_point.distance(point)
                < snap_threshold * snap_threshold_error_multiplier
                for intersecting_point in node_candidates.geometry.values
            ]
            assert all(isinstance(val, bool) for val in intersection_data)
            # If no intersects for current node -> continue to next
            if sum(intersection_data) == 0:
                continue

            # Different error thresholds for endpoints and intersections
            if sum(intersection_data) >= error_threshold:
                # Add idx of current trace
                indexes_with_junctions.add(idx)
                # Add other point idxs that correspond to the error
                for other_index in node_candidates.loc[
                    intersection_data
                ].index.to_list():  # type: ignore
                    indexes_with_junctions.add(flattened_idx_reference[other_index])

    return indexes_with_junctions


def zip_equal(*iterables):
    """
    Zip iterables of only equal lengths.
    """
    sentinel = object()
    for combo in zip_longest(*iterables, fillvalue=sentinel):
        if sentinel in combo:
            raise ValueError("Iterables have different lengths")
        yield combo


def create_unit_vector(start_point: Point, end_point: Point) -> np.ndarray:
    """
    Create numpy unit vector from two shapely Points.

    :param start_point: The start point.
    :param end_point: The end point.
    :return: The unit vector that points from ``start_point`` to
        ``end_point``.
    """
    # Convert Point coordinates to (x, y)
    segment_start = point_to_xy(start_point)
    segment_end = point_to_xy(end_point)
    segment_vector = np.array(
        [segment_end[0] - segment_start[0], segment_end[1] - segment_start[1]]
    )
    if any(np.isnan(segment_vector)):
        return np.array([np.nan, np.nan])
    segment_unit_vector = segment_vector / np.linalg.norm(segment_vector)
    assert isinstance(segment_unit_vector, np.ndarray)
    return segment_unit_vector


def compare_unit_vector_orientation(
    vec_1: np.ndarray, vec_2: np.ndarray, threshold_angle: float
) -> bool:
    """
    If vec_1 and vec_2 are too different in orientation, will return False.
    """
    if np.linalg.norm(vec_1 + vec_2) < np.sqrt(2):
        # If they face opposite side -> False
        return False
    dot_product = np.dot(vec_1, vec_2)
    if np.isclose(dot_product, 1):
        return True
    if dot_product > 1 or dot_product < -1 or np.isnan(dot_product):
        return False
    rad_angle = np.arccos(dot_product)
    deg_angle = np.rad2deg(rad_angle)
    if deg_angle > threshold_angle:
        # If angle between more than threshold_angle -> False
        return False
    return True


def bounding_polygon(geoseries: Union[gpd.GeoSeries, gpd.GeoDataFrame]) -> Polygon:
    """
    Create bounding polygon around GeoSeries.

    The geoseries geometries will always be completely enveloped by the
    polygon. The geometries will not intersect the polygon boundary.

    >>> geom = LineString([(1, 0), (1, 1), (-1, -1)])
    >>> geoseries = gpd.GeoSeries([geom])
    >>> poly = bounding_polygon(geoseries)
    >>> poly.wkt
    'POLYGON ((2 -2, 2 2, -2 2, -2 -2, 2 -2))'
    >>> geoseries.intersects(poly.boundary)
    0    False
    dtype: bool

    """
    total_bounds_geoseries = total_bounds(geoseries)
    bounding_poly: Polygon = scale(box(*total_bounds_geoseries), xfact=2, yfact=2)
    assert isinstance(bounding_poly, Polygon)
    if any(
        geom.intersects(bounding_poly.boundary) for geom in geoseries.geometry.values
    ):
        bounding_poly = safe_buffer(bounding_poly, radius=1.0)
        if any(
            geom.intersects(bounding_poly.boundary)
            for geom in geoseries.geometry.values
        ):
            # if any(geoseries.intersects(bounding_polygon.boundary)):
            raise ValueError("Expected no intersects after buffer.")
    assert all(geoseries.within(bounding_poly))
    return bounding_poly


def mls_to_ls(multilinestrings: List[MultiLineString]) -> List[LineString]:
    """
    Flattens a list of multilinestrings to a list of linestrings.

    >>> multilinestrings = [
    ...     MultiLineString(
    ...         [
    ...             LineString([(1, 1), (2, 2), (3, 3)]),
    ...             LineString([(1.9999, 2), (-2, 5)]),
    ...         ]
    ...     ),
    ...     MultiLineString(
    ...         [
    ...             LineString([(1, 1), (2, 2), (3, 3)]),
    ...             LineString([(1.9999, 2), (-2, 5)]),
    ...         ]
    ...     ),
    ... ]
    >>> result_linestrings = mls_to_ls(multilinestrings)
    >>> print([ls.wkt for ls in result_linestrings])
    ['LINESTRING (1 1, 2 2, 3 3)', 'LINESTRING (1.9999 2, -2 5)',
    'LINESTRING (1 1, 2 2, 3 3)', 'LINESTRING (1.9999 2, -2 5)']

    """
    linestrings: List[LineString] = []
    for mls in multilinestrings:
        linestrings.extend(list(mls.geoms))
    if not all(isinstance(ls, LineString) for ls in linestrings):
        raise ValueError("MultiLineStrings within MultiLineStrings?")
    return linestrings


def efficient_clip(
    traces: Union[gpd.GeoSeries, gpd.GeoDataFrame],
    areas: Union[gpd.GeoSeries, gpd.GeoDataFrame],
) -> gpd.GeoDataFrame:
    """
    Perform efficient clip of LineString geometries with a Polygon.

    :param traces: Trace data.
    :param areas: Area data.
    :return: Traces clipped with the area data.
    """
    # Transform to pygeos types
    pygeos_traces = pygeos.from_shapely(traces.geometry.values)

    # Convert MultiPolygon in area_gdf to Polygons and collect to list.
    polygons = []
    for geom in areas.geometry.values:
        if isinstance(geom, MultiPolygon):
            polygons.extend(geom.geoms)
        elif isinstance(geom, Polygon):
            polygons.append(geom)
        else:
            raise TypeError(
                f"Expected (Multi)Polygons in efficient_clip."
                f" Got: {geom.wkt, type(geom)}."
            )
    pygeos_polygons = pygeos.from_shapely(polygons)
    pygeos_multipolygon = pygeos.multipolygons(pygeos_polygons)

    # Perform intersection
    intersection = pygeos.intersection(pygeos_traces, pygeos_multipolygon)
    assert isinstance(intersection, np.ndarray)

    # Collect into GeoDataFrame.
    geodataframe = gpd.GeoDataFrame(geometry=intersection, crs=traces.crs)
    assert "geometry" in geodataframe.columns
    return geodataframe


@JOBLIB_CACHE.cache
def crop_to_target_areas(
    traces: Union[gpd.GeoSeries, gpd.GeoDataFrame],
    areas: Union[gpd.GeoSeries, gpd.GeoDataFrame],
    is_filtered: bool = False,
    keep_column_data: bool = False,
    allow_multilinestring_input: bool = False,
) -> Union[gpd.GeoSeries, gpd.GeoDataFrame]:
    """
    Crop traces to the area polygons.

    E.g.

    >>> traces = gpd.GeoSeries(
    ...     [LineString([(1, 1), (2, 2), (3, 3)]), LineString([(1.9999, 2), (-2, 5)])]
    ... )
    >>> areas = gpd.GeoSeries(
    ...     [
    ...         Polygon([(1, 1), (-1, 1), (-1, -1), (1, -1)]),
    ...         Polygon([(-2.5, 6), (-1.9, 6), (-1.9, 4), (-2.5, 4)]),
    ...     ]
    ... )
    >>> cropped_traces = crop_to_target_areas(traces, areas, 0.01)
    >>> print([trace.wkt for trace in cropped_traces.geometry.values])
    ['LINESTRING (-1.9 4.924998124953124, -2 5)']

    :param traces: Trace data.
    :param areas: Area data.
    :param snap_threshold: Distance threshold for snapping.
    :param is_filtered: Has preliminary spatial filtering already been
        done. If not the traces will be filtered first with a spatial index.
    :param keep_column_data: Is column data of traces required to persist.
    :return: Cropped traces.
    :raises TypeError: If all geometries in the end result are not
        ``LineString's``.
    """
    # Only handle LineStrings
    if (
        not all(isinstance(trace, LineString) for trace in traces.geometry.values)
        and not allow_multilinestring_input
    ):
        raise TypeError(
            "Expected no MultiLineString geometries in crop_to_target_areas."
        )
    # Match the crs
    traces, areas = match_crs(traces, areas)

    if not is_filtered:
        traces.reset_index(drop=True, inplace=True)
        spatial_index = pygeos_spatial_index(traces)

        areas_bounds = total_bounds(areas)
        assert len(areas_bounds) == 4

        candidate_idxs = spatial_index_intersection(
            spatial_index=spatial_index, coordinates=areas_bounds
        )
        candidate_traces: Union[gpd.GeoSeries, gpd.GeoDataFrame] = traces.iloc[
            candidate_idxs
        ]
    else:
        candidate_traces = traces

    # TODO: Remove environment check
    if keep_column_data or os.environ.get("USE_PYGEOS") == "0":
        assert all(area_geom.is_valid for area_geom in areas.geometry.values)
        assert all(not area_geom.is_empty for area_geom in areas.geometry.values)
        # geopandas.clip keeps the column data
        try:
            clipped_traces = gpd.clip(candidate_traces, areas)
        except TypeError:
            logging.error(
                "Expected to be able to clip with geopandas. "
                "Falling back to pygeos clip.",
                exc_info=True,
            )
            clipped_traces = efficient_clip(candidate_traces, areas)
    else:
        # pygeos.intersection does not
        clipped_traces = efficient_clip(candidate_traces, areas)

    assert hasattr(clipped_traces, "geometry")
    assert isinstance(clipped_traces, (gpd.GeoDataFrame, gpd.GeoSeries))

    # Debug log for traces smaller than MINIMUM_LINE_LENGTH
    sum_smaller_than_minimum = sum(clipped_traces.geometry.length < MINIMUM_LINE_LENGTH)
    if sum_smaller_than_minimum > 0:
        # Log if found
        log.info(
            "Traces smaller than MINIMUM_LINE_LENGTH found after crop.",
            extra=dict(
                MINIMUM_LINE_LENGTH=MINIMUM_LINE_LENGTH,
                found_amount=sum_smaller_than_minimum,
            ),
        )

    # Clipping might result in Point geometries
    # Filter to only LineStrings and MultiLineStrings
    clipped_traces = clipped_traces.loc[
        [
            isinstance(geom, (LineString, MultiLineString))
            for geom in clipped_traces.geometry.values
        ]
    ]

    # Some traces might be converted to MultiLineStrings if they become
    # disjointed with the clip.
    # This presents a data management problem -> Data is either lost or
    # duplicated to different parts of a MultiLineString converted trace
    clipped_and_dissolved_traces = dissolve_multi_part_traces(clipped_traces)
    assert (
        clipped_and_dissolved_traces.shape[0] >= clipped_and_dissolved_traces.shape[0]
    )

    # Remove small traces that might cause topology errors
    # Filtering out traces smaller in length than
    # fractopo.general.MINIMUM_LINE_LENGTH
    clipped_and_dissolved_traces = clipped_and_dissolved_traces.loc[
        clipped_and_dissolved_traces.geometry.length > MINIMUM_LINE_LENGTH
    ]
    log.info(
        "Filtered out small traces with MINIMUM_LINE_LENGTH.",
        extra=dict(
            MINIMUM_LINE_LENGTH=MINIMUM_LINE_LENGTH,
            new_minimum=clipped_and_dissolved_traces.geometry.length.min(),
            maximum=clipped_and_dissolved_traces.geometry.length.max(),
        ),
    )

    # clipped_and_dissolved_traces = clipped_and_dissolved_traces.loc[
    #     clipped_and_dissolved_traces.geometry.length > snap_threshold * 2.01
    # ]

    # Reset index
    clipped_and_dissolved_traces.reset_index(inplace=True, drop=True)

    return clipped_and_dissolved_traces


@overload
def dissolve_multi_part_traces(traces: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """
    Overload for gpd.GeoDataFrame.
    """
    ...


@overload
def dissolve_multi_part_traces(traces: gpd.GeoSeries) -> gpd.GeoSeries:
    """
    Overload for gpd.GeoSeries.
    """
    ...


def dissolve_multi_part_traces(
    traces: Union[gpd.GeoDataFrame, gpd.GeoSeries]
) -> Union[gpd.GeoDataFrame, gpd.GeoSeries]:
    """
    Dissolve MultiLineStrings in GeoDataFrame or GeoSeries.

    Copies all attribute data of rows with MultiLineStrings to new LineStrings.
    """
    # Get MultiLineString rows
    mls_bools = [isinstance(trace, MultiLineString) for trace in traces.geometry.values]
    mls_traces = traces.loc[
        [isinstance(trace, MultiLineString) for trace in traces.geometry.values]
    ]
    # If no MultiLineString geoms -> return original
    if mls_traces.shape[0] == 0:
        return traces

    # Gather LineString geoms
    ls_traces = traces.loc[[not val for val in mls_bools]]

    # Dissolve MultiLineString geoms but keep same row data
    dissolved_rows = []

    as_linestrings_list = [mls_to_ls([geom]) for geom in mls_traces.geometry.values]
    if isinstance(traces, gpd.GeoSeries):
        return gpd.GeoSeries(list(chain(*as_linestrings_list)), crs=traces.crs)

    for (_, row), as_linestrings in zip(mls_traces.iterrows(), as_linestrings_list):
        # as_linestrings = mls_to_ls([row.geometry])
        if len(as_linestrings) == 0:
            raise ValueError("Expected at least one geom from mls_to_ls.")

        for new_geom in as_linestrings:
            # Do not modify mls_traces inplace
            new_row = row.copy()
            new_row[GEOMETRY_COLUMN] = new_geom
            dissolved_rows.append(new_row)

    # Merge with ls_traces
    dissolved_rows_gdf = gpd.GeoDataFrame(dissolved_rows, crs=ls_traces.crs)
    assert ls_traces.crs == dissolved_rows_gdf.crs
    dissolved_traces = pd.concat([ls_traces, dissolved_rows_gdf])
    assert isinstance(dissolved_traces, gpd.GeoDataFrame)

    if not all(isinstance(val, LineString) for val in dissolved_traces.geometry.values):
        raise TypeError("Expected all LineStrings in dissolved_traces.")
    return dissolved_traces


def is_empty_area(area: gpd.GeoDataFrame, traces: gpd.GeoDataFrame):
    """
    Check if any traces intersect the area(s) in area GeoDataFrame.
    """
    for area_polygon in area.geometry.values:
        sindex = pygeos_spatial_index(traces)
        intersection = spatial_index_intersection(sindex, geom_bounds(area_polygon))
        potential_traces = traces.iloc[intersection]

        for trace in potential_traces.geometry.values:
            # Only one trace intersect required. No need to loop
            # through all traces if one found.
            if trace.intersects(area_polygon):
                return False
    return True


def resolve_split_to_ls(geom: LineString, splitter: LineString) -> List[LineString]:
    """
    Resolve split between two LineStrings to only LineString results.
    """
    split_current = list(split(geom, splitter).geoms)
    linestrings = [
        geom
        for geom in split_current
        if isinstance(geom, LineString) and not geom.is_empty
    ]
    linestrings.extend(
        mls_to_ls([geom for geom in split_current if isinstance(geom, MultiLineString)])
    )
    return linestrings


def safe_buffer(
    geom: Union[Point, LineString, Polygon], radius: float, **kwargs
) -> Polygon:
    """
    Get type checked Polygon buffer.

    >>> result = safe_buffer(Point(0, 0), 1)
    >>> isinstance(result, Polygon), round(result.area, 3)
    (True, 3.137)
    """
    buffer = geom.buffer(radius, **kwargs)
    if not isinstance(buffer, Polygon):
        raise TypeError("Expected Polygon buffer.")
    return buffer


def random_points_within(poly: Polygon, num_points: int) -> List[Point]:
    """
    Get random points within Polygon.

    >>> from pprint import pprint
    >>> random.seed(10)
    >>> poly = box(0, 0, 1, 1)
    >>> result = random_points_within(poly, 2)
    >>> pprint([point.within(poly) for point in result])
    [True, True]
    """
    min_x, min_y, max_x, max_y = geom_bounds(poly)
    points: List[Point] = []

    while len(points) < num_points:
        random_point = Point(
            [random.uniform(min_x, max_x), random.uniform(min_y, max_y)]
        )
        if random_point.within(poly):
            points.append(random_point)

    return points


def spatial_index_intersection(
    spatial_index: PyGEOSSTRTreeIndex, coordinates: Union[BoundsTuple, PointTuple]
) -> List[int]:
    """
    Type-checked spatial index intersection.

    TODO: Maybe use spatial_index.quary_bulk for faster execution?
    TODO: Maybe use a predicate (e.g., "contains") to specify operation?
    """
    if spatial_index is None:
        return []
    result = spatial_index.intersection(coordinates)
    indexes = []
    for idx in result:
        if isinstance(idx, int):
            indexes.append(idx)
        elif idx == int(idx):
            indexes.append(idx)
        else:
            raise TypeError("Expected integer results from intersection.")
    return indexes


def within_bounds(
    x: float, y: float, min_x: float, min_y: float, max_x: float, max_y: float
):
    """
    Are x and y within the bounds.

    >>> within_bounds(1, 1, 0, 0, 2, 2)
    True
    """
    return (min_x <= x <= max_x) and (min_y <= y <= max_y)


def geom_bounds(
    geom: Union[LineString, Polygon, MultiPolygon]
) -> Tuple[float, float, float, float]:
    """
    Get LineString or Polygon bounds.

    >>> geom_bounds(LineString([(-10, -10), (10, 10)]))
    (-10.0, -10.0, 10.0, 10.0)
    """
    types = (LineString, Polygon, MultiPolygon)
    if not isinstance(geom, types):
        raise TypeError(f"Expected {types} as geom type.")
    bounds = []
    for val in geom.bounds:
        if isinstance(val, (int, float)):
            bounds.append(val)
        else:
            raise TypeError("Expected numerical bounds.")
    if len(bounds) != 4:
        raise ValueError(f"Expected bounds of length 4. Got: {bounds}.")
    bounds_tuple: Tuple[float, float, float, float] = (
        bounds[0],
        bounds[1],
        bounds[2],
        bounds[3],
    )
    return bounds_tuple


def total_bounds(
    geodata: Union[gpd.GeoSeries, gpd.GeoDataFrame]
) -> Tuple[float, float, float, float]:
    """
    Get total bounds of geodataset.

    >>> geodata = gpd.GeoSeries([Point(-10, 10), Point(10, 10)])
    >>> total_bounds(geodata)
    (-10.0, 10.0, 10.0, 10.0)
    """
    if geodata.empty or all(
        geom is None or geom.is_empty for geom in geodata.geometry.values
    ):
        return tuple([np.nan] * 4)
    bounds = geodata.total_bounds
    if not len(bounds) == 4:
        raise ValueError(
            f"Expected total_bounds to return an array of length 4: {bounds}."
        )
    return bounds[0], bounds[1], bounds[2], bounds[3]


def pygeos_spatial_index(
    geodataset: Union[gpd.GeoDataFrame, gpd.GeoSeries]
) -> PyGEOSSTRTreeIndex:
    """
    Get PyGEOSSTRTreeIndex from geopandas dataset.

    :param geodataset: Geodataset of which
        spatial index is wanted.
    :return: ``pygeos`` spatial index.
    :raises TypeError: If the geodataset ``sindex`` attribute was not a
        ``pygeos`` spatial index object.
    """
    sindex = geodataset.sindex
    if not isinstance(sindex, PyGEOSSTRTreeIndex):
        raise TypeError("Expected PyGEOSSTRTreeIndex as spatial index.")
    return sindex


def read_geofile(path: Path) -> gpd.GeoDataFrame:
    """
    Read a filepath for a ``GeoDataFrame`` representable geo-object.

    :param path: ``pathlib.Path`` to a ``GeoDataFrame``
        representable spatial file.
    :return: ``geopandas.GeoDataFrame`` read from the file.
    :raises TypeError: If the file could not be parsed as a ``GeoDataFrame``
        by ``geopandas``.
    """
    data = gpd.read_file(path)
    if not isinstance(data, gpd.GeoDataFrame):
        raise TypeError("Expected GeoDataFrame as file read result.")
    return data


def determine_boundary_intersecting_lines(
    line_gdf: gpd.GeoDataFrame, area_gdf: gpd.GeoDataFrame, snap_threshold: float
):
    """
    Determine lines that intersect any target area boundary.
    """
    assert isinstance(line_gdf, (gpd.GeoSeries, gpd.GeoDataFrame))
    # line_gdf = line_gdf.reset_index(inplace=False, drop=True)
    # spatial_index = line_gdf.sindex
    spatial_index = pygeos_spatial_index(line_gdf)
    intersecting_idxs = []
    cuts_through_idxs = []
    for target_area in area_gdf.geometry.values:
        assert isinstance(target_area, (MultiPolygon, Polygon))
        # target_area_bounds = target_area.boundary.bounds
        target_area_bounds = geom_bounds(target_area)
        assert len(target_area_bounds) == 4 and isinstance(target_area_bounds, tuple)
        min_x, min_y, max_x, max_y = target_area_bounds
        # TODO: Use spatial_index.quary_bulk for faster execution?
        # TODO: Use a predicate (e.g., "contains") to specify operation?
        intersection = spatial_index.intersection(
            extend_bounds(
                min_x=min_x,
                min_y=min_y,
                max_x=max_x,
                max_y=max_y,
                extend_amount=snap_threshold * 100,
            )
        )
        candidate_idxs = list(intersection if intersection is not None else [])
        if len(candidate_idxs) == 0:
            continue

        for candidate_idx in candidate_idxs:
            line = line_gdf.iloc[candidate_idx].geometry
            assert isinstance(line, LineString)
            if line.distance(target_area.boundary) < snap_threshold:
                intersecting_idxs.append(candidate_idx)
                endpoints = get_trace_endpoints(line)
                if all(
                    endpoint.distance(target_area.boundary) < snap_threshold
                    for endpoint in endpoints
                ):
                    cuts_through_idxs.append(candidate_idx)

                elif not any(
                    endpoint.within(target_area) for endpoint in endpoints
                ) and np.isclose(line.distance(target_area), 0):
                    cuts_through_idxs.append(candidate_idx)

    intersecting_lines = np.array(
        [idx in intersecting_idxs for idx in line_gdf.index.values]
    )
    cuts_through_lines = np.array(
        [idx in cuts_through_idxs for idx in line_gdf.index.values]
    )
    assert intersecting_lines.dtype == "bool"
    assert cuts_through_lines.dtype == "bool"
    return intersecting_lines, cuts_through_lines


def extend_bounds(
    min_x: float, min_y: float, max_x: float, max_y: float, extend_amount: float
) -> Tuple[float, float, float, float]:
    """
    Extend bounds by addition and reduction.

    >>> extend_bounds(0, 0, 10, 10, 10)
    (-10, -10, 20, 20)
    """
    return (
        min_x - extend_amount,
        min_y - extend_amount,
        max_x + extend_amount,
        max_y + extend_amount,
    )


def bool_arrays_sum(arr_1: np.ndarray, arr_2: np.ndarray) -> np.ndarray:
    """
    Calculate integer sum of two arrays.

    Resulting array consists only of integers 0, 1 and 2.

    >>> arr_1 = np.array([True, False, False])
    >>> arr_2 = np.array([True, True, False])
    >>> bool_arrays_sum(arr_1, arr_2)
    array([2, 1, 0])

    >>> arr_1 = np.array([True, True])
    >>> arr_2 = np.array([True, True])
    >>> bool_arrays_sum(arr_1, arr_2)
    array([2, 2])
    """
    assert arr_1.dtype == "bool"
    assert arr_2.dtype == "bool"
    return np.array([int(val_1) + int(val_2) for val_1, val_2 in zip(arr_1, arr_2)])


def intersection_count_to_boundary_weight(intersection_count: int) -> int:
    """
    Get actual weight factor for boundary intersection count.

    >>> intersection_count_to_boundary_weight(2)
    0
    >>> intersection_count_to_boundary_weight(0)
    1
    >>> intersection_count_to_boundary_weight(1)
    2
    """
    if not isinstance(intersection_count, int):
        intersection_count = intersection_count.item()
    assert isinstance(intersection_count, int)
    if intersection_count == 0:
        return 1
    if intersection_count == 1:
        return 2
    if intersection_count == 2:
        return 0
    raise ValueError(f"Expected 0,1,2 as intersection_count. Got: {intersection_count}")


def numpy_to_python_type(value: Any):
    """
    Convert numpy dtype variable to Python type, if possible.
    """
    try:
        return value.item()
    except AttributeError:
        return value


def calc_circle_area(radius: float) -> float:
    """
    Calculate area of circle.

    >>> calc_circle_area(1.78)
    9.953822163633902
    """
    return np.pi * radius**2


def calc_circle_radius(area: float) -> float:
    """
    Calculate radius from area.

    >>> calc_circle_radius(10.0)
    1.7841241161527712

    """
    assert not area < 0
    radius = numpy_to_python_type(np.sqrt(area / np.pi))
    assert isinstance(radius, float)
    return radius


def point_to_point_unit_vector(point: Point, other_point: Point) -> np.ndarray:
    """
    Create unit vector from point to other point.

    >>> point = Point(0, 0)
    >>> other_point = Point(1, 1)
    >>> point_to_point_unit_vector(point, other_point)
    array([0.70710678, 0.70710678])
    """
    x1, y1 = point.x, point.y
    x2, y2 = other_point.x, other_point.y
    if any(np.isnan([x1, y1, x2, y2])):
        raise ValueError(f"Expected no nan values: {x1, y1, x2, y2}")
    vector = np.array([x2 - x1, y2 - y1])
    if np.isclose(point.distance(other_point), 0.0, rtol=1e-15, atol=1e-15):
        logging.info(
            "Cannot compute unit vector between points due to close approximity."
        )
        return np.array([np.nan, np.nan])
    normed_vector = vector / np.linalg.norm(vector)
    assert isinstance(normed_vector, np.ndarray)
    return normed_vector


def raise_determination_error(
    attribute: str,
    verb: str = "determining",
    determine_target: str = "topology (=branches and nodes)",
):
    """
    Raise AttributeError if attribute cannot be determined.

    >>> try:
    ...     raise_determination_error("parameters")
    ...     assert False
    ... except AttributeError as exc:
    ...     print(f"{str(exc)[0:20]}...")
    ...
    Cannot determine par...

    """
    raise AttributeError(
        f"Cannot determine {attribute} without {verb} {determine_target}."
    )


def multiprocess(
    function_to_call: Callable,
    keyword_arguments: Sequence,
    arguments_identifier=lambda _: "",
    repeats: int = 0,
) -> List[ProcessResult]:
    """
    Process function calls in parallel.

    Returns result as a list where the error is appended when execution fails.
    """
    # Collect results into a list
    collect_results: List[ProcessResult] = []

    # multiprocessing!
    with ProcessPoolExecutor() as executor:
        keyword_arguments = list(keyword_arguments) * (repeats + 1)

        # Iterate over invalids. submit as tasks
        futures = {
            executor.submit(function_to_call, keyword_arg): keyword_arg
            for keyword_arg in keyword_arguments
        }

        # Collect all tasks as they complete
        # Will not be in same order as submitted
        for future in as_completed(futures):
            identifier = arguments_identifier(futures[future])
            # If execution critically fails it will be caught and logged
            try:
                # Get result from Future
                # This will throw an error (if it happened in process)
                result = future.result()

                # Collect result
                collect_results.append(
                    ProcessResult(identifier=identifier, result=result, error=False)
                )
            except Exception as exc:
                # Catch and log critical failures
                log.error(f"Process exception with {futures[future]}:\n\n" f"{exc}")
                collect_results.append(
                    ProcessResult(identifier=identifier, error=True, result=exc)
                )
    return collect_results


def save_fig(fig: Figure, results_dir: Path, name: str) -> List[Path]:
    """
    Save figure as svg image to results dir.
    """
    assert results_dir.exists() and results_dir.is_dir()
    assert len(name) > 0
    output_paths = []
    for suffix in ("png", "svg"):
        output_path = results_dir / f"{name}.{suffix}"
        fig.savefig(output_path, bbox_inches="tight")
        output_paths.append(output_path)
    return output_paths


@contextmanager
def silent_output(name: str):
    """
    General method to silence output from general func.
    """
    # Create memory files for stdout and stderr
    tmp_io_stdout = StringIO()
    tmp_io_stderr = StringIO()

    # Use double context managers and yield within both
    try:
        with redirect_stdout(tmp_io_stdout):
            with redirect_stderr(tmp_io_stderr):
                yield

    # Report stdout and stderr to log.info
    finally:
        tmp_io_stdout_value = tmp_io_stdout.getvalue()
        tmp_io_stderr_value = tmp_io_stderr.getvalue()
        # Do not log without output
        if len(tmp_io_stdout_value) + len(tmp_io_stderr_value) > 5:
            log.info(
                "silent_output output stdout and stderr.",
                extra=dict(
                    func_name=name,
                    stdout=tmp_io_stdout_value,
                    stderr=tmp_io_stderr_value,
                ),
            )


def wrap_silence(func):
    """
    Wrap function to capture and silence its output.

    Used primarily to silence output from ``powerlaw`` functions and methods.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        with silent_output(func.__name__):
            results = func(*args, **kwargs)
        return results

    return wrapper


def convert_list_columns(gdf: gpd.GeoDataFrame, allow: bool = True) -> gpd.GeoDataFrame:
    """
    Convert list type columns to string.
    """
    gdf = gdf.copy()
    if gdf.empty:
        return gdf
    for column in gdf.columns.values:
        column_data = gdf[column]
        assert isinstance(column_data, pd.Series)
        if isinstance(column_data.values[0], list):
            if not allow:
                raise ValueError(
                    "Trying to write geodata to disk but it has columns with"
                    "`list`-type data and `allow` is set to `False`.\n"
                    "Either set allow to True or fix the data."
                )
            log.info(f"Converting {column} from list to str.")
            gdf[column] = [str(tuple(item)) for item in column_data.values]
    return gdf


def write_geodata(
    gdf: gpd.GeoDataFrame,
    path: Path,
    driver: str = GEOJSON_DRIVER,
    allow_list_column_transform: bool = False,
):
    """
    Write geodata with driver.

    Default is GeoJSON.
    """
    if gdf.empty:
        # Handle empty GeoDataFrames
        path.write_text(gdf.to_json(sort_keys=True))
    else:
        # Convert list type columns to string.
        gdf = convert_list_columns(gdf, allow=allow_list_column_transform)

        # Write to disk
        gdf.to_file(path, driver=driver)

    # Only geojson output is formatted afterwards.
    if driver != GEOJSON_DRIVER:
        return

    # Format geojson with indent of 1
    read_json = path.read_text()
    loaded_json = json.loads(read_json)
    dumped_json = json.dumps(loaded_json, indent=1, sort_keys=True)
    path.write_text(dumped_json)


def write_topodata(gdf: gpd.GeoDataFrame, output_path: Path):
    """
    Write branch or nodes GeoDataFrame to `output_path`.
    """
    write_geodata(gdf=gdf, path=output_path, allow_list_column_transform=False)


def azimuth_to_unit_vector(azimuth: float) -> np.ndarray:
    """
    Convert azimuth to unit vector.
    """
    azimuth_rad = np.deg2rad(azimuth)
    return np.array([np.sin(azimuth_rad), np.cos(azimuth_rad)])


def r2_scorer(y_true: np.ndarray, y_predicted: np.ndarray) -> float:
    """
    Score fit with r2 metric.

    Changes the scoring to be best at a value of 0.
    """
    score = abs(1 - sklm.r2_score(y_true=y_true, y_pred=y_predicted))
    assert isinstance(score, float)
    return score


def focus_plot_to_bounds(ax: Axes, total_bounds: np.ndarray) -> Axes:
    """
    Focus plot to given bounds.
    """
    xmin, ymin, xmax, ymax = total_bounds
    extend_x = (xmax - xmin) * 0.05
    extend_y = (ymax - ymin) * 0.05
    ax.set_xlim(xmin - extend_x, xmax + extend_x)
    ax.set_ylim(ymin - extend_y, ymax + extend_y)
    return ax


def assign_branch_and_node_colors(feature_type: str) -> str:
    """
    Determine color for each branch and node type.

    Defaults to red.

    >>> assign_branch_and_node_colors("C-C")
    'red'
    """
    if feature_type in (CC_branch, X_node):
        return "green"
    if feature_type in (CI_branch, Y_node):
        return "blue"
    if feature_type in (II_branch, I_node):
        return "black"
    return "red"


def remove_duplicate_caseinsensitive_columns(columns) -> set:
    """
    Remove duplicate columns case-insensitively.
    """
    lower_case_columns = set(column.lower() for column in columns)
    new_cols = set(columns)
    for column in columns:
        if column not in lower_case_columns and column.lower() in columns:
            new_cols.remove(column)
    return new_cols


def write_geodataframe(geodataframe: gpd.GeoDataFrame, name: str, results_dir: Path):
    """
    Save geodataframe as GeoPackage, GeoJSON and shapefile.
    """
    non_dupl_columns = remove_duplicate_caseinsensitive_columns(geodataframe.columns)
    fid_col = "fid"
    for column in geodataframe.columns:
        if column not in non_dupl_columns:
            # print(f"Dropping column: {column}")
            geodataframe.drop(columns=[column], inplace=True)
        if column.lower() == fid_col:
            # Remove fid columns
            # print(f"Dropping column: {column} due to case-insensitive match to fid.")
            geodataframe.drop(columns=[column], inplace=True)
    error_base = "Failed to write {} as {}."
    try:
        geodataframe.to_file(results_dir / f"{name}.gpkg", driver="GPKG")
    except Exception:
        log.error(error_base.format(name, "GeoPackage"), exc_info=True)
    try:
        geodataframe.to_file(results_dir / f"{name}.geojson", driver="GeoJSON")
    except Exception:
        log.error(error_base.format(name, "GeoJSON"), exc_info=True)
    try:
        shp_dir = results_dir / f"{name}_as_shp"
        shp_dir.mkdir(exist_ok=False)
        geodataframe.to_file(shp_dir / f"{name}.shp")
    except Exception:
        log.error(error_base.format(name, "ESRI Shapefile"), exc_info=True)


def check_for_z_coordinates(geodata: Union[gpd.GeoDataFrame, gpd.GeoSeries]) -> bool:
    """
    Check if geopandas data contains Z-coordinates.
    """
    return any(geom.has_z for geom in geodata.geometry.values if hasattr(geom, "has_z"))


def remove_z_coordinates_from_geodata(
    geodata: Union[gpd.GeoDataFrame, gpd.GeoSeries]
) -> Union[gpd.GeoDataFrame, gpd.GeoSeries]:
    """
    Remove Z-coordinates from geometries in geopandasgeodata.
    """
    geodata_without_z = geodata.copy()
    geodata_without_z_geometries = geodata_without_z.geometry.apply(
        remove_z_coordinates
    )
    if isinstance(geodata_without_z, gpd.GeoDataFrame):
        geodata_without_z["geometry"] = geodata_without_z_geometries
    else:
        geodata_without_z = geodata_without_z_geometries

    assert isinstance(geodata_without_z, (gpd.GeoDataFrame, gpd.GeoSeries))
    return geodata_without_z


def remove_z_coordinates(geometry: Union[BaseGeometry, BaseMultipartGeometry]) -> Any:
    """
    Remove z-coordinates from a geometry.

    A ``shapely`` geometry should be provided. Output will always be the same
    type of geometry with the z-coordinates removed.

    :param geometry: Shapely geometry
    :return: Shapely geometry with the same geometry type
    """
    return wkb.loads(wkb.dumps(geometry, output_dimension=2))


def sanitize_name(name: str) -> str:
    """
    Return only alphanumeric parts of name string.
    """
    return "".join(filter(str.isalnum, name))


def check_for_wrong_geometries(traces: gpd.GeoDataFrame, area: gpd.GeoDataFrame):
    """
    Check that traces are line geometries and area contains area geometries.
    """
    accepted_line_types = (LineString, MultiLineString)
    accepted_area_types = (Polygon, MultiPolygon)
    for name, geoms, accepted_types in zip(
        ("traces", "area"), (traces, area), (accepted_line_types, accepted_area_types)
    ):
        if not all(isinstance(geom, accepted_types) for geom in geoms.geometry.values):
            raise TypeError(f"Expected {name} to contain only any of {accepted_types}.")



"""
Utilities for randomly Network sampling traces.
"""
import logging
from dataclasses import dataclass
from enum import Enum, unique
from typing import Optional, Tuple, Union

import geopandas as gpd
import numpy as np
from shapely.geometry import LineString, Point, Polygon

from fractopo.analysis.network import Network
from fractopo.general import (
    GEOMETRY_COLUMN,
    calc_circle_area,
    calc_circle_radius,
    numpy_to_python_type,
    random_points_within,
    safe_buffer,
)

log = logging.getLogger(__name__)


@unique
class RandomChoice(Enum):

    """
    Choose between random area or radius.

    The choice is relevant because area is polynomially correlated with radius.
    """

    area = "area"
    radius = "radius"


@dataclass
class RandomSample:

    """
    Dataclass for sampling results.
    """

    network_maybe: Optional[Network]
    target_centroid: Point
    radius: float
    name: str


@dataclass
class NetworkRandomSampler:

    """
    Randomly sample traces inside given target area.
    """

    trace_gdf: gpd.GeoDataFrame
    area_gdf: gpd.GeoDataFrame
    min_radius: float
    snap_threshold: float
    random_choice: Union[RandomChoice, str]
    name: str

    def __post_init__(self):
        """
        Validate inputs post-initialization.
        """
        self.random_choice = self.random_choice_should_be_enum(self.random_choice)
        self.trace_gdf = self.trace_gdf_should_contain_traces(self.trace_gdf)
        self.area_gdf = self.area_gdf_should_contain_polygon(self.area_gdf)
        self.min_radius = self.value_should_be_positive(self.min_radius)

    @staticmethod
    def random_choice_should_be_enum(
        random_choice: Union[RandomChoice, str]
    ) -> RandomChoice:
        """
        Check that random_choice is valid.
        """
        # Convert strings if they match any enum values
        if isinstance(random_choice, str):
            for choice_enum in RandomChoice:
                if random_choice == choice_enum.value:
                    return choice_enum

            raise TypeError(
                "Expected random_choice to be"
                f" convertible to RandomChoice enum: {random_choice}."
            )
        return random_choice

    @staticmethod
    def trace_gdf_should_contain_traces(
        trace_gdf: gpd.GeoDataFrame,
    ) -> gpd.GeoDataFrame:
        """
        Check that trace_gdf contains LineString traces.
        """
        if not trace_gdf.shape[0] > 0:
            raise ValueError("Expected non-empty trace_gdf.")
        if not all(isinstance(val, LineString) for val in trace_gdf.geometry.values):
            raise TypeError("Expected only LineStrings in trace_gdf.")
        return trace_gdf

    @staticmethod
    def area_gdf_should_contain_polygon(area_gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """
        Check that area_gdf contains one Polygon.
        """
        if not area_gdf.shape[0] == 1:
            raise ValueError("Expected area_gdf with one geometry.")
        geom = area_gdf.geometry.values[0]
        if not isinstance(geom, Polygon):
            raise TypeError("Expected one Polygon in area_gdf.")
        return area_gdf

    @staticmethod
    def value_should_be_positive(min_radius: float) -> float:
        """
        Check that value is positive.
        """
        if not min_radius > 0:
            raise ValueError(f"Expected positive non-zero min_radius: {min_radius}.")
        return min_radius

    @property
    def target_circle(self) -> Polygon:
        """
        Target circle Polygon from area_gdf.
        """
        return self.area_gdf.geometry.values[0]

    @property
    def max_radius(self) -> float:
        """
        Calculate max radius from given area_gdf.
        """
        radius = numpy_to_python_type(np.sqrt(self.target_circle.area / np.pi))
        assert isinstance(radius, float)
        if self.min_radius > radius:
            raise ValueError("Expected min_radius smaller than max_radius.")
        return radius

    @property
    def min_area(self) -> float:
        """
        Calculate minimum area from min_radius.
        """
        return calc_circle_area(self.min_radius)

    @property
    def max_area(self) -> float:
        """
        Calculate maximum area from max_radius.
        """
        return calc_circle_area(self.max_radius)

    def random_radius(self) -> float:
        """
        Calculate random radius in range [min_radius, max_radius[.
        """
        radius_range = self.max_radius - self.min_radius
        radius = self.min_radius + np.random.random_sample() * radius_range
        return radius

    def random_area(self) -> float:
        """
        Calculate random area in area range.

        Range is calculated from [min_radius, max_radius[.
        """
        area_range = self.max_area - self.min_area
        area = self.min_area + np.random.random_sample() * area_range
        return area

    @property
    def target_area_centroid(self) -> Point:
        """
        Get target area centroid.
        """
        centroid = self.target_circle.centroid
        if not isinstance(centroid, Point):
            raise TypeError("Expected Point as centroid.")
        return centroid

    def random_target_circle(self) -> Tuple[Polygon, Point, float]:
        """
        Get random target area and its centroid and radius.

        The target area is always within the original target area.
        """
        radius = (
            self.random_radius()
            if self.random_choice == RandomChoice.radius
            else calc_circle_radius(self.random_area())
        )
        possible_buffer: Polygon = safe_buffer(
            self.target_area_centroid, self.max_radius - radius
        )
        random_target_centroid = random_points_within(possible_buffer, 1)[0]
        random_target_circle: Polygon = safe_buffer(random_target_centroid, radius)

        return random_target_circle, random_target_centroid, radius

    def random_network_sample(self, determine_branches_nodes=True) -> RandomSample:
        """
        Get random Network sample with a random target area.

        Returns the network, the sample circle centroid and circle radius.
        """
        # Create random Polygon circle within area
        target_circle, target_centroid, radius = self.random_target_circle()

        # Collect into GeoDataFrame and set crs if it exists in input frame
        area_gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [target_circle]})
        if self.trace_gdf.crs is not None:
            area_gdf = area_gdf.set_crs(self.trace_gdf.crs)

        try:
            network_maybe: Optional[Network] = Network(
                trace_gdf=self.trace_gdf,
                area_gdf=area_gdf,
                name=self.name,
                determine_branches_nodes=determine_branches_nodes,
                snap_threshold=self.snap_threshold,
                circular_target_area=True,
                truncate_traces=True,
            )
        except ValueError:
            log.error(
                "Error occurred during creation of random_network_sample.",
                exc_info=True,
                extra=dict(
                    sampler_name=self.name,
                    target_centroid=target_centroid,
                    radius=radius,
                ),
            )
            network_maybe = None

        return RandomSample(
            network_maybe=network_maybe,
            target_centroid=target_centroid,
            radius=radius,
            name=self.name,
        )

    @classmethod
    def random_network_sampler(
        cls,
        network: Network,
        min_radius: float,
        random_choice: RandomChoice = RandomChoice.radius,
    ):
        """
        Initialize ``NetworkRandomSampler`` for random sampling.

        Assumes that ``Network`` target area is a single ``Polygon`` circle.
        """
        if not network.circular_target_area:
            raise ValueError(
                "Expected passed ``Network`` to be initialized with"
                " circular_target_area as ``True``."
            )
        return NetworkRandomSampler(
            trace_gdf=network.trace_gdf,
            area_gdf=network.area_gdf,
            min_radius=min_radius,
            snap_threshold=network.snap_threshold,
            random_choice=random_choice,
            name=network.name,
        )



"""
Analysis and plotting of geometric and topological parameters.
"""
import logging
from itertools import compress
from textwrap import fill
from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import ternary
from matplotlib import patheffects as path_effects
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from scipy.ndimage import gaussian_filter
from ternary.ternary_axes_subplot import TernaryAxesSubplot

from fractopo.general import (
    CC_branch,
    CE_branch,
    CI_branch,
    E_node,
    EE_branch,
    I_node,
    IE_branch,
    II_branch,
    Number,
    Param,
    X_node,
    Y_node,
)

log = logging.getLogger(__name__)


def determine_node_type_counts(
    node_types: np.ndarray, branches_defined: bool
) -> Dict[str, Number]:
    """
    Determine node type counts.
    """
    return {
        str(node_class): (
            (sum(node_types == node_class) if len(node_types) > 0 else 0)
            if branches_defined
            else np.nan
        )
        for node_class in (X_node, Y_node, I_node, E_node)
    }


def determine_branch_type_counts(
    branch_types: np.ndarray, branches_defined: bool
) -> Dict[str, Number]:
    """
    Determine branch type counts.
    """
    return {
        str(branch_class): sum(branch_types == branch_class)
        if branches_defined
        else np.nan
        for branch_class in (
            CC_branch,
            CI_branch,
            II_branch,
            CE_branch,
            IE_branch,
            EE_branch,
        )
    }


def ternary_text(text: str, ax: Axes):
    """
    Add ternary text about counts.
    """
    prop = dict(boxstyle="square", facecolor="linen", alpha=1, pad=0.45)
    ax.text(
        0.85,
        1.05,
        text,
        transform=ax.transAxes,
        fontsize="medium",
        weight="roman",
        verticalalignment="top",
        bbox=prop,
        fontfamily="DejaVu Sans",
        ha="center",
    )


def decorate_xyi_ax(ax: Axes, tax: TernaryAxesSubplot, counts: Dict[str, int]):
    """
    Decorate xyi plot.
    """
    xcount, ycount, icount = _get_xyi_counts(counts)
    text = "\n".join(
        (
            f"n: {xcount+ycount+icount}",
            f"X-nodes: {xcount}",
            f"Y-nodes: {ycount}",
            f"I-nodes: {icount}",
        )
    )
    initialize_ternary_points(ax, tax)
    tern_plot_the_fing_lines(tax)
    ternary_text(ax=ax, text=text)


def decorate_count_ax(
    ax: Axes, tax: TernaryAxesSubplot, label_counts: Dict[str, int], is_nodes: bool
):
    """
    Decorate ternary count plot.
    """
    text = "\n".join([f"{label} (n={value})" for label, value in label_counts.items()])
    if is_nodes:
        initialize_ternary_points(ax=ax, tax=tax)
        tern_plot_the_fing_lines(tax)
    else:
        initialize_ternary_branches_points(ax=ax, tax=tax)
        tern_plot_branch_lines(tax=tax)
    prop = dict(boxstyle="square", facecolor="linen", alpha=1, pad=0.45)
    ax.text(
        0.85,
        1.05,
        text,
        transform=ax.transAxes,
        fontsize="medium",
        weight="roman",
        verticalalignment="top",
        bbox=prop,
        fontfamily="DejaVu Sans",
        ha="center",
    )


def branches_intersect_boundary(branch_types: np.ndarray) -> np.ndarray:
    """
    Get array of if branches have E-component (intersects target area).
    """
    array_isin = ~np.isin(branch_types, (CC_branch, CI_branch, II_branch))
    assert isinstance(array_isin, np.ndarray)
    return array_isin


def plot_ternary_plot(
    counts_list: List[Dict[str, int]],
    labels: List[str],
    is_nodes: bool,
    colors: Optional[List[Optional[str]]] = None,
) -> Tuple[Figure, Axes, TernaryAxesSubplot]:
    """
    Plot ternary plot.

    Same function is used to plot both XYI and branch type ternary plots.
    """
    # Check if plotting xyi or branches
    # Set plotter and decorator functions accordingly
    if is_nodes:
        plotter = plot_xyi_plot_ax
        decorator = decorate_xyi_ax
    else:
        plotter = plot_branch_plot_ax
        decorator = decorate_branch_ax

    # Check if only one label is given
    one_label = len(labels) == 1

    # Check if colors are given and if they are, check that there are enough
    # for all ternary points
    if colors is None:
        colors = [None for _ in counts_list]
    elif len(colors) != len(counts_list):
        raise ValueError(
            f"Expected colors (len={len(colors)}) to be of"
            f" same size as node_counts_list (len={len(counts_list)})."
        )

    # Scale in the plots is 100
    scale = 100

    # Initialize ternary plot
    fig, ax = plt.subplots(figsize=(6.5, 5.1))
    fig, tax = ternary.figure(ax=ax, scale=scale)

    # If only one set of counts is give, we are only plotting a single ternary
    # point
    if len(counts_list) == 1:
        # Call plotter with single point
        plotter(counts=counts_list[0], label=labels[0], tax=tax, color=colors[0])

        # Decorate plot
        decorator(ax, tax, counts=counts_list[0])
    else:
        # Gather points from tuples of counts
        points = [
            counts_to_point(counts=count, scale=100, is_nodes=is_nodes)
            for count in counts_list
        ]

        # Filter to X,Y,I or CC,CI,II using get_counts_func
        get_counts_func = _get_xyi_counts if is_nodes else _get_branch_class_counts
        if one_label:
            counts_sum = sum(sum(get_counts_func(count)) for count in counts_list)
            label_counts = {"n": counts_sum}
        else:
            label_counts = {
                label: sum(get_counts_func(count))
                for label, count in zip(labels, counts_list)
            }
        decorate_count_ax(ax=ax, tax=tax, label_counts=label_counts, is_nodes=is_nodes)

        # Create boolean list of if point is not None
        points_is_not_none = [point is not None for point in points]

        # Remove None-points
        points = list(compress(points, points_is_not_none))

        if one_label:
            # Only one label for all points is given
            tax.scatter(points, color="black", **ternary_point_kwargs(s=35, zorder=3))
            tax.scatter(
                points,
                label=labels[0],
                color=colors[0],
                **ternary_point_kwargs(),
            )
        else:
            # Multiple labels are given
            labels = list(compress(labels, points_is_not_none))
            colors = list(compress(colors, points_is_not_none))
            assert len(points) == len(labels)
            assert len(points) == len(colors)
            for point, label, color in zip(points, labels, colors):
                # Plot individually while giving unique label for each
                tax.scatter(
                    [point], color="black", **ternary_point_kwargs(s=35, zorder=3)
                )
                tax.scatter([point], label=label, color=color, **ternary_point_kwargs())
    tax.legend(
        loc="upper center",
        bbox_to_anchor=(0.5, -0.15),
        prop={"family": "DejaVu Sans", "weight": "heavy", "size": "x-large"},
        edgecolor="black",
        ncol=2,
        columnspacing=0.7,
        shadow=True,
    )

    # Return Figure, Axes and TernaryAxesSubplot
    return fig, ax, tax


def ternary_heatmapping(
    x_values: np.ndarray,
    y_values: np.ndarray,
    i_values: np.ndarray,
    number_of_bins: int,
    scale_divider: float = 1.0,
    ax: Optional[Axes] = None,
) -> Tuple[Figure, ternary.TernaryAxesSubplot]:
    """
    Plot ternary heatmap.

    Modified from: https://github.com/marcharper/python-ternary/issues/81
    """
    scale = number_of_bins / scale_divider
    histogram, _ = np.histogramdd(
        (x_values, y_values, i_values),
        bins=(number_of_bins, number_of_bins, number_of_bins),
        range=((0, 1), (0, 1), (0, 1)),
    )
    histogram_normed = histogram / np.sum(histogram)

    # 3D smoothing and interpolation
    kde = gaussian_filter(histogram_normed, sigma=2)
    interp_dict = dict()

    binx = np.linspace(0, 1, number_of_bins)
    for i in range(len(binx)):
        for j in range(len(binx)):
            for k in range(len(binx)):
                interp_dict[(i, j, k)] = kde[i, j, k]

    fig, tax = ternary.figure(ax=ax, scale=scale)
    tax.heatmap(interp_dict)
    return fig, tax


def counts_to_point(
    counts: Dict[str, int], is_nodes: bool, scale: int = 100
) -> Optional[Tuple[float, float, float]]:
    """
    Create ternary point from node_counts.

    The order is important: for nodes: X, I, Y and for branches: CC, II, CI.
    """
    if is_nodes:
        # xcount, ycount, icount = _get_xyi_counts(node_counts)
        point_counts = _get_xyi_counts(counts)
    else:
        # cc_count, ci_count, ii_count = _get_branch_class_counts(branch_counts)
        point_counts = _get_branch_class_counts(counts)

    sumcount = sum(point_counts)
    if sumcount == 0 or np.isclose(sumcount, 0.0):
        return None
    # NOTE: Order is switched to get right order in output.
    first = scale * point_counts[0] / sumcount
    second = scale * point_counts[2] / sumcount
    third = scale * point_counts[1] / sumcount
    return (first, second, third)


def _get_xyi_counts(node_counts: Dict[str, int]) -> Tuple[int, int, int]:
    """
    Return tuple of node counts from dict of node counts.
    """
    xcount = node_counts[X_node]
    ycount = node_counts[Y_node]
    icount = node_counts[I_node]
    return xcount, ycount, icount


def _get_branch_class_counts(branch_counts: Dict[str, int]) -> Tuple[int, int, int]:
    """
    Return tuple of branch counts from dict of branch counts.
    """
    cc_count = branch_counts[CC_branch]
    ci_count = branch_counts[CI_branch]
    ii_count = branch_counts[II_branch]
    return cc_count, ci_count, ii_count


def plot_xyi_plot_ax(
    counts: Dict[str, int],
    label: str,
    tax: ternary.ternary_axes_subplot.TernaryAxesSubplot,
    color: Optional[str] = None,
):
    """
    Plot XYI pointst to given ternary axis (tax).
    """
    if color is None:
        color = "black"
    # xcount, ycount, icount = _get_xyi_counts(node_counts)
    point = counts_to_point(counts, scale=100, is_nodes=True)
    if point is not None:
        # plot_ternary_point(tax=tax, point=point, marker="o", label=label, color=color)
        # Add shadow to point
        tax.scatter(
            points=[point],
            **ternary_point_kwargs(marker="o", s=35),
            color="black",
        )
        tax.scatter(
            points=[point], **ternary_point_kwargs(marker="o"), label=label, color=color
        )
    tax.legend(
        loc="upper center",
        bbox_to_anchor=(0.5, -0.15),
        prop={"family": "DejaVu Sans", "weight": "heavy", "size": "x-large"},
        edgecolor="black",
        ncol=2,
        columnspacing=0.7,
        shadow=True,
    )


def ternary_point_kwargs(
    alpha=1.0,
    zorder=4,
    s: float = 25,
    marker="X",
):
    """
    Plot point to a ternary figure.
    """
    return dict(
        alpha=alpha,
        zorder=zorder,
        s=s,
        marker=marker,
    )


def plot_branch_plot_ax(
    counts: Dict[str, int],
    label: str,
    tax: ternary.ternary_axes_subplot.TernaryAxesSubplot,
    color: Optional[str] = None,
):
    """
    Plot ternary branch plot to tax.
    """
    if color is None:
        color = "black"
    # cc_count, ci_count, ii_count = _get_branch_class_counts(branch_counts)
    point = counts_to_point(counts, is_nodes=False)
    if point is not None:
        tax.scatter(
            points=[point],
            **ternary_point_kwargs(marker="o", s=35, zorder=3),
            color="black",
        )
        tax.scatter(
            points=[point],
            **ternary_point_kwargs(marker="o"),
            label=label,
            color=color,
        )

        # plot_ternary_point(tax=tax, point=point, marker="o", label=label, color=color)
    tax.legend(
        loc="upper center",
        bbox_to_anchor=(0.5, -0.15),
        prop={"family": "DejaVu Sans", "weight": "heavy", "size": "x-large"},
        edgecolor="black",
        ncol=2,
        columnspacing=0.7,
        shadow=True,
    )


def decorate_branch_ax(
    ax: Axes,
    tax: TernaryAxesSubplot,
    # label: str,
    counts: Dict[str, int],
):
    """
    Decorate ternary branch plot.
    """
    cc_count, ci_count, ii_count = _get_branch_class_counts(counts)
    text = "\n".join(
        (
            f"n: {cc_count+ci_count+ii_count}",
            f"CC-branches: {cc_count}",
            f"CI-branches: {ci_count}",
            f"II-branches: {ii_count}",
        )
    )
    # prop = dict(boxstyle="square", facecolor="linen", alpha=1, pad=0.45)
    # ax.text(
    #     0.86,
    #     1.05,
    #     text,
    #     transform=ax.transAxes,
    #     fontsize="medium",
    #     weight="roman",
    #     verticalalignment="top",
    #     bbox=prop,
    #     fontfamily="DejaVu Sans",
    #     ha="center",
    # )
    initialize_ternary_branches_points(ax, tax)
    tern_plot_branch_lines(tax)
    ternary_text(ax=ax, text=text)
    # tax.legend(
    #     loc="upper center",
    #     bbox_to_anchor=(0.1, 1.05),
    #     prop={"family": "DejaVu Sans", "weight": "heavy", "size": "x-large"},
    #     edgecolor="black",
    #     ncol=2,
    #     columnspacing=0.7,
    #     shadow=True,
    # )


def determine_topology_parameters(
    trace_length_array: np.ndarray,
    area: float,
    branches_defined: bool,
    correct_mauldon: bool,
    node_counts: Optional[Dict[str, Number]],
    branch_length_array: Optional[np.ndarray],
) -> Dict[str, float]:
    """
    Determine geometric (and topological) parameters.

    Number of traces (and branches) are determined by node counting.

    The passed ``trace_length_array`` should be non-weighted.
    """
    radius = np.sqrt(area / np.pi)
    if len(trace_length_array) == 0:
        min_length_traces, max_length_traces, characteristic_length_traces = (
            0.0,
            0.0,
            0.0,
        )
    else:
        min_length_traces, max_length_traces, characteristic_length_traces = (
            float(trace_length_array.min()),
            float(trace_length_array.max()),
            float(trace_length_array.mean()),
        )
    assert isinstance(min_length_traces, float), type(min_length_traces)
    assert isinstance(max_length_traces, float)
    assert isinstance(characteristic_length_traces, float)
    # characteristic_length_traces = (
    #     trace_length_array.mean() if len(trace_length_array) > 0 else 0.0
    # )
    fracture_intensity = trace_length_array.sum() / area
    dimensionless_intensity_traces = fracture_intensity * characteristic_length_traces

    # Collect parameters that do not require topology determination into a dict
    params_without_topology = {
        Param.FRACTURE_INTENSITY_B21.value.name: fracture_intensity,
        Param.FRACTURE_INTENSITY_P21.value.name: fracture_intensity,
        Param.TRACE_MIN_LENGTH.value.name: min_length_traces,
        Param.TRACE_MAX_LENGTH.value.name: max_length_traces,
        Param.TRACE_MEAN_LENGTH.value.name: characteristic_length_traces,
        Param.DIMENSIONLESS_INTENSITY_P22.value.name: dimensionless_intensity_traces,
        Param.AREA.value.name: area,
        Param.NUMBER_OF_TRACES_TRUE.value.name: len(trace_length_array),
    }

    if not branches_defined:
        # Return dict with nans in place of topological parameters
        nan_dict = {
            param.value.name: np.nan
            for param in Param
            if param.value.name not in params_without_topology
        }
        all_params_without_topo = {**params_without_topology, **nan_dict}
        assert all(param.value.name in all_params_without_topo for param in Param)

        return all_params_without_topo

    assert isinstance(trace_length_array, np.ndarray)
    assert isinstance(node_counts, dict)
    assert isinstance(area, float)

    if branch_length_array is None:
        raise ValueError("Expected branch_length_array to not be None.")

    number_of_traces = (node_counts[Y_node] + node_counts[I_node]) / 2

    # Handle empty branch length array
    if len(branch_length_array) == 0:
        min_length_branches, max_length_branches = 0.0, 0.0
    else:
        min_length_branches, max_length_branches = (
            float(branch_length_array.min()),
            float(branch_length_array.max()),
        )
    aerial_frequency_traces = number_of_traces / area
    number_of_branches = (
        (node_counts[X_node] * 4) + (node_counts[Y_node] * 3) + node_counts[I_node]
    ) / 2
    aerial_frequency_branches = number_of_branches / area
    characteristic_length_branches = (
        (trace_length_array.sum() / number_of_branches)
        if number_of_branches > 0
        else 0.0
    )
    dimensionless_intensity_branches = (
        fracture_intensity * characteristic_length_branches
    )
    connections_per_trace = (
        (2 * (node_counts[Y_node] + node_counts[X_node]) / number_of_traces)
        if number_of_traces > 0
        else 0.0
    )
    connections_per_branch = (
        (3 * node_counts[Y_node] + 4 * node_counts[X_node]) / number_of_branches
        if number_of_branches > 0
        else 0.0
    )
    if correct_mauldon:
        trace_mean_length_mauldon = (
            (
                ((np.pi * radius) / 2)
                * (node_counts[E_node] / (node_counts[I_node] + node_counts[Y_node]))
            )
            if node_counts[I_node] + node_counts[Y_node] > 0
            else 0.0
        )
        fracture_density_mauldon = (node_counts[I_node] + node_counts[Y_node]) / (
            area * 2
        )
        fracture_intensity_mauldon = (
            trace_mean_length_mauldon * fracture_density_mauldon
        )
    else:
        # If Network target area is not circular mauldon parameters cannot be
        # determined.
        (
            trace_mean_length_mauldon,
            fracture_density_mauldon,
            fracture_intensity_mauldon,
        ) = (np.nan, np.nan, np.nan)
    connection_frequency = (node_counts[Y_node] + node_counts[X_node]) / area

    params_with_topology = {
        Param.NUMBER_OF_TRACES.value.name: number_of_traces,
        Param.BRANCH_MIN_LENGTH.value.name: min_length_branches,
        Param.BRANCH_MAX_LENGTH.value.name: max_length_branches,
        Param.BRANCH_MEAN_LENGTH.value.name: characteristic_length_branches,
        Param.AREAL_FREQUENCY_B20.value.name: aerial_frequency_branches,
        Param.AREAL_FREQUENCY_P20.value.name: aerial_frequency_traces,
        Param.DIMENSIONLESS_INTENSITY_B22.value.name: dimensionless_intensity_branches,
        Param.CONNECTIONS_PER_TRACE.value.name: connections_per_trace,
        Param.CONNECTIONS_PER_BRANCH.value.name: connections_per_branch,
        Param.FRACTURE_INTENSITY_MAULDON.value.name: fracture_intensity_mauldon,
        Param.FRACTURE_DENSITY_MAULDON.value.name: fracture_density_mauldon,
        Param.TRACE_MEAN_LENGTH_MAULDON.value.name: trace_mean_length_mauldon,
        Param.CONNECTION_FREQUENCY.value.name: connection_frequency,
        Param.NUMBER_OF_BRANCHES.value.name: number_of_branches,
        Param.NUMBER_OF_BRANCHES_TRUE.value.name: len(branch_length_array),
    }

    all_parameters = {**params_without_topology, **params_with_topology}
    assert len(all_parameters) == sum(
        [len(params_without_topology), len(params_with_topology)]
    )
    assert all(
        param in {param.value.name for param in Param} for param in all_parameters
    )

    return all_parameters


def plot_parameters_plot(
    topology_parameters_list: List[Dict[str, float]],
    labels: List[str],
    colors: Optional[List[str]] = None,
):
    """
    Plot topological parameters.
    """
    prop = dict(boxstyle="square", facecolor="linen", alpha=1, pad=0.45)

    columns_to_plot = [
        param for param in Param if param.value.name in topology_parameters_list[0]
    ]
    figs, axes = [], []

    width = 6 + 1 * len(topology_parameters_list) / 6
    bar_width = 0.6 * len(topology_parameters_list) / 6

    topology_concat = pd.DataFrame(topology_parameters_list)
    topology_concat["label"] = labels
    topology_concat.label = topology_concat.label.astype("category")

    for column in columns_to_plot:
        fig, ax = plt.subplots(figsize=(width, 5.5))

        if all(np.isnan(topology_concat[column.value.name])):
            figs.append(fig)
            axes.append(ax)
            continue

        # Figure size setup
        # TODO: width higher, MAYBE lower bar_width

        # Trying to have sensible widths for bars:
        topology_concat.plot.bar(
            x="label",
            y=column.value.name,
            color=colors,
            zorder=5,
            alpha=0.9,
            width=bar_width,
            ax=ax,
        )
        # PLOT STYLING
        ax.set_xlabel("")
        ax.set_ylabel(
            column.value.name + " " + f"({column.value.unit})",
            fontsize="xx-large",
            fontfamily="DejaVu Sans",
            style="italic",
        )
        ax.set_title(
            x=0.5,
            y=1.09,
            label=column.value.name,
            fontsize="xx-large",
            bbox=prop,
            transform=ax.transAxes,
            zorder=-10,
        )
        legend = ax.legend()
        legend.remove()
        if column.value.plot_as_log and not all(
            topology_concat[column.value.name] <= 0
        ):
            ax.set_yscale("log")
        fig.subplots_adjust(top=0.85, bottom=0.25, left=0.2)
        locs, xtick_labels = plt.xticks()
        # xtick_labels = ["\n".join(wrap(label.get_text(), 6)) for label in xtick_labels]
        xtick_labels = [fill(label.get_text(), 6) for label in xtick_labels]
        plt.yticks(fontsize="xx-large", color="black")
        plt.xticks(locs, labels, fontsize="xx-large", color="black")
        # VALUES ABOVE BARS WITH TEXTS
        rects = ax.patches
        for value, rect in zip(topology_concat[column.value.name].values, rects):
            height = rect.get_height()
            if value > 0.01:
                value = round(value, 2)
            else:
                value = f"{value:.2e}"
            if column.value.plot_as_log:
                height = height + height / 10
            else:
                max_height = max(r.get_height() for r in rects)
                height = height + max_height / 100
            ax.text(
                rect.get_x() + rect.get_width() / 2,
                height,
                value,
                ha="center",
                va="bottom",
                zorder=15,
                fontsize="x-large",
            )
        figs.append(fig)
        axes.append(ax)
    return figs, axes


def determine_set_counts(
    set_names: Tuple[str, ...], set_array: np.ndarray
) -> Dict[str, int]:
    """
    Determine counts in for each set.
    """
    return {
        set_name: sum(set_array == set_name)
        if sum(set_array == set_name) is not None
        else 0
        for set_name in set_names
    }


def plot_set_count(
    set_counts: Dict[str, int],
    label: str,
) -> Tuple[Figure, Axes]:
    """
    Plot set counts.
    """
    fig, ax = plt.subplots(figsize=(7, 7))

    log_info = dict(set_counts=set_counts, label=label)

    # Handle case with 0 values for all keys in set_counts dict.
    if all(value == 0 for value in set_counts.values()):
        log.warning(
            "set_counts with only 0 values passed into plot_set_count.",
            extra=log_info,
        )
        return fig, ax

    # Report creation.
    log.info(
        "Creating pie plot for set counts.",
        extra=log_info,
    )

    # Create the plot
    _, label_texts, _ = ax.pie(
        x=[set_counts[key] for key in set_counts],
        labels=list(set_counts),
        autopct="%.1f%%",
        explode=[0.025 for _ in set_counts],
        pctdistance=0.5,
        textprops=dict(weight="bold"),
        wedgeprops=dict(linewidth=2, edgecolor="black"),
    )

    # Set title
    ax.set_title(label)
    for label_text in label_texts:
        label_text.set_fontsize("large")
    return fig, ax


def initialize_ternary_points(ax: Axes, tax: TernaryAxesSubplot):
    """
    Initialize ternary points figure ax and tax.
    """
    fdict = initialize_ternary_ax(ax=ax, tax=tax)
    ax.text(-0.1, -0.03, "Y", transform=ax.transAxes, fontdict=fdict)
    ax.text(1.03, -0.03, "X", transform=ax.transAxes, fontdict=fdict)
    ax.text(0.5, 1.07, "I", transform=ax.transAxes, fontdict=fdict, ha="center")


def tern_plot_the_fing_lines(tax: TernaryAxesSubplot, cs_locs=(1.3, 1.5, 1.7, 1.9)):
    """
    Plot *connections per branch* parameter to XYI-plot.

    If not using the pre-determined locations the texts will not be correctly
    placed as they use absolute positions and labels.

    :param tax: Ternary axis to plot to.
    :param cs_locs: Pre-determined locations for the lines.
    """

    def tern_find_last_x(c, x_start=0):
        x, _, y = tern_yi_func(c, x_start)
        while y > 0:
            x_start += 0.01
            x, _, y = tern_yi_func(c, x_start)
        return x

    def tern_yi_func_perc(c, x):
        temp = 6 * (1 - 0.5 * c)
        temp2 = 3 - (3 / 2) * c
        temp3 = 1 + c / temp
        y = (c + 3 * c * x) / (temp * temp3) - (4 * x) / (temp2 * temp3)
        i = 1 - x - y
        return 100 * x, 100 * i, 100 * y

    for c in cs_locs:
        last_x = tern_find_last_x(c)
        x1 = 0
        x2 = last_x
        point1 = tern_yi_func_perc(c, x1)
        point2 = tern_yi_func_perc(c, x2)
        tax.line(
            point1,
            point2,
            alpha=0.6,
            color="k",
            zorder=-5,
            linestyle="dashed",
            linewidth=0.75,
        )
        ax = plt.gca()
        rot = 6.5
        rot2 = 4.5
        ax.text(x=55, y=59, s=r"$C_B = 1.3$", fontsize=10, rotation=rot, ha="center")
        ax.text(x=61, y=50, s=r"$C_B = 1.5$", fontsize=10, rotation=rot, ha="center")
        ax.text(
            x=68.5,
            y=36.6,
            s=r"$C_B = 1.7$",
            fontsize=10,
            rotation=rot2 + 1,
            ha="center",
        )
        ax.text(x=76, y=17, s=r"$C_B = 1.9$", fontsize=10, rotation=rot2, ha="center")


def initialize_ternary_ax(ax: Axes, tax: TernaryAxesSubplot) -> dict:
    """
    Decorate ternary ax for both XYI and branch types.

    Returns a font dict for use in plotting text.
    """
    # Remove normal matplotlib axes
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)

    # Remove frame
    ax.set_frame_on(False)

    # Set ternary boundary
    tax.boundary(linewidth=1.5)

    # Add gridlines in the tax
    tax.gridlines(linewidth=0.15, multiple=20, color="black", alpha=0.6, zorder=-100)

    # Add ticks
    tax.ticks(
        axis="lbr",
        linewidth=0.9,
        multiple=20,
        offset=0.03,
        tick_formats="%d%%",
        fontsize="small",
        alpha=0.6,
    )

    # Clear the default matplotlib ticks
    tax.clear_matplotlib_ticks()

    # Another check to remove matplotlib axes
    tax.get_axes().axis("off")

    # Create font dict for use in plotting text
    fontsize = 20
    fdict = {
        "path_effects": [path_effects.withStroke(linewidth=3, foreground="k")],
        "color": "w",
        "family": "DejaVu Sans",
        "size": fontsize,
        "weight": "bold",
    }
    return fdict


def initialize_ternary_branches_points(ax, tax):
    """
    Initialize ternary branches plot ax and tax.
    """
    fdict = initialize_ternary_ax(ax=ax, tax=tax)
    ax.text(-0.1, -0.06, "I - C", transform=ax.transAxes, fontdict=fdict)
    ax.text(1.0, -0.06, "C - C", transform=ax.transAxes, fontdict=fdict)
    ax.text(0.5, 1.07, "I - I", transform=ax.transAxes, fontdict=fdict, ha="center")


def tern_plot_branch_lines(tax):
    """
    Plot line of random assignment of nodes to branches to a branch ternary tax.

    Line positions taken from NetworkGT open source code.
    Credit to:
    https://github.com/BjornNyberg/NetworkGT

    :param tax: Ternary axis to plot to
    :type tax: ternary.TernaryAxesSubplot
    """
    ax = tax.get_axes()
    tax.boundary()
    points = [
        (0, 1, 0),
        (0.01, 0.81, 0.18),
        (0.04, 0.64, 0.32),
        (0.09, 0.49, 0.42),
        (0.16, 0.36, 0.48),
        (0.25, 0.25, 0.5),
        (0.36, 0.16, 0.48),
        (0.49, 0.09, 0.42),
        (0.64, 0.04, 0.32),
        (0.81, 0.01, 0.18),
        (1, 0, 0),
    ]
    for idx, _ in enumerate(points):
        points[idx] = points[idx][0] * 100, points[idx][1] * 100, points[idx][2] * 100

    text_loc = [(0.37, 0.2), (0.44, 0.15), (0.52, 0.088), (0.64, 0.055), (0.79, 0.027)]
    for idx, t in enumerate(text_loc):
        text_loc[idx] = t[0] * 100, t[1] * 100
    text = [r"$C_B = 1.0$", r"$1.2$", r"$1.4$", r"$1.6$", r"$1.8$"]
    rots = [-61, -44, -28, -14, -3]
    # rot = -65
    for t, l, rot in zip(text, text_loc, rots):
        ax.annotate(t, xy=l, fontsize=9, rotation=rot)
        # rot += 17
    tax.plot(
        points,
        linewidth=1.5,
        marker="o",
        color="k",
        linestyle="dashed",
        markersize=3,
        zorder=-5,
        alpha=0.6,
    )


def tern_yi_func(c, x):
    """
    Plot *Connections per branch* threshold line to branch ternary plot.

    Uses absolute values.
    """
    temp = 6 * (1 - 0.5 * c)
    temp2 = 3 - (3 / 2) * c
    temp3 = 1 + c / temp
    y = (c + 3 * c * x) / (temp * temp3) - (4 * x) / (temp2 * temp3)
    i = 1 - x - y
    return x, i, y


def convert_counts(counts: Dict[str, Number]) -> Dict[str, int]:
    """
    Convert float and int value in counts to ints only.

    Used by branches and nodes count calculators.
    """
    if any(np.isnan(list(counts.values()))):
        raise ValueError(f"Expected no nan in counts: {counts}")
    counts_ints = {
        key: int(value) for key, value in counts.items() if not np.isnan(value)
    }
    if len(counts_ints) != len(counts):
        raise ValueError(f"Expected no nan in node_counts: {counts}")

    assert all(isinstance(value, int) for value in counts_ints.values())
    return counts_ints



"""
MultiNetwork implementation for handling multiple network analysis.
"""

from typing import Dict, List, NamedTuple, Optional, Tuple, Type, Union

import pandas as pd
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from fractopo.analysis import length_distributions, parameters, subsampling
from fractopo.analysis.network import Network
from fractopo.analysis.random_sampling import RandomChoice
from fractopo.general import NAME, Number, SetRangeTuple


class MultiNetwork(NamedTuple):

    """
    Multiple Network analysis.
    """

    networks: Tuple[Network, ...]

    def __hash__(self) -> int:
        """
        Implement hashing for MultiNetwork.
        """
        return hash(tuple(self.networks))

    def subsample(
        self,
        min_radii: Union[float, Dict[str, float]],
        random_choice: RandomChoice = RandomChoice.radius,
        samples: int = 1,
    ) -> List[Optional[Dict[str, Union[Number, str]]]]:
        """
        Subsample all ``Networks`` in ``MultiNetwork``.

        :param min_radii: The minimum radius for
            subsamples can be either given as static or as a mapping that maps
            network names to specific radii.
        :param random_choice: Whether to use radius or area
            as the random choice parameter.
        :param samples: How many subsamples to conduct per network.
        :returns: Subsamples
        """
        return subsampling.subsample_networks(
            networks=self.networks,
            min_radii=min_radii,
            random_choice=random_choice,
            samples=samples,
        )

    def collective_azimuth_sets(self) -> Tuple[Tuple[str, ...], SetRangeTuple]:
        """
        Get collective azimuth set names.

        Checks that all Networks have the same azimuth sets.
        """
        # Check that set_names and set_ranges are the same across all Networks
        set_names = self.networks[0].azimuth_set_names
        set_ranges = self.networks[0].azimuth_set_ranges
        for network in self.networks:
            if (
                set_names != network.azimuth_set_names
                and set_ranges == network.azimuth_set_ranges
            ):
                raise ValueError(
                    "Expected azimuth sets and ranges to be the same"
                    " in all networks."
                )
        return set_names, set_ranges

    def network_length_distributions(
        self, using_branches: bool, using_azimuth_sets: bool
    ) -> Dict[str, Dict[str, length_distributions.LengthDistribution]]:
        """
        Get length distributions of Networks.
        """
        # Get azimuth sets that are identical for all Networks.
        set_names, _ = self.collective_azimuth_sets()

        # Collect (set-wise) length distributions for all networks
        distributions: Dict[
            str, Dict[str, length_distributions.LengthDistribution]
        ] = dict()

        # Iterate over the networks
        for network in self.networks:
            # Name is either the network name or if categorizing by set,
            # the set name
            # Only one LengthDistribution is collected if not using the set,
            # otherwise an amount equal to the amount of azimuth sets
            names = [network.name] if not using_azimuth_sets else set_names

            # Collect length distributions (for each set)
            network_distributions = dict()
            for name in names:
                # Label the LengthDistribution with either just network name
                # or with the name and the azimuth set
                azimuth_set = name if using_azimuth_sets else None

                # # Use network name or set name as key in second dictionary
                network_distributions[name] = (
                    network.branch_length_distribution(azimuth_set=azimuth_set)
                    # Use either trace or branch lengths
                    if using_branches
                    else network.trace_length_distribution(azimuth_set=azimuth_set)
                )

            # Use network name as key
            distributions[network.name] = network_distributions
        return distributions

    def multi_length_distributions(
        self, using_branches: bool
    ) -> length_distributions.MultiLengthDistribution:
        """
        Get MultiLengthDistribution of all networks.
        """
        distributions_dict = self.network_length_distributions(
            using_branches=using_branches, using_azimuth_sets=False
        )
        distributions = [
            distributions_dict[network.name][network.name] for network in self.networks
        ]

        multi_distribution = length_distributions.MultiLengthDistribution(
            distributions=distributions,
            using_branches=using_branches,
        )
        return multi_distribution

    def plot_multi_length_distribution(
        self,
        using_branches: bool,
        automatic_cut_offs: bool,
        plot_truncated_data: bool,
        multi_distribution: Optional[
            length_distributions.MultiLengthDistribution
        ] = None,
    ) -> Tuple[
        length_distributions.MultiLengthDistribution,
        length_distributions.Polyfit,
        Figure,
        Axes,
    ]:
        """
        Plot multi-length distribution fit.

        Use ``multi_length_distributions()`` to get most parameters used in
        fitting the multi-scale distribution.
        """
        multi_distribution = (
            self.multi_length_distributions(using_branches=using_branches)
            if multi_distribution is None
            else multi_distribution
        )
        polyfit, fig, ax = multi_distribution.plot_multi_length_distributions(
            automatic_cut_offs=automatic_cut_offs,
            plot_truncated_data=plot_truncated_data,
        )

        return multi_distribution, polyfit, fig, ax

    def plot_xyi(
        self,
        colors: Optional[List[Optional[str]]] = None,
    ):
        """
        Plot multi-network ternary XYI plot.
        """
        node_counts_list = [network.node_counts for network in self.networks]
        labels = [network.name for network in self.networks]
        return parameters.plot_ternary_plot(
            counts_list=node_counts_list,
            labels=labels,
            is_nodes=True,
            colors=colors,
        )

    def plot_branch(
        self,
        colors: Optional[List[Optional[str]]] = None,
    ):
        """
        Plot multi-network ternary branch type plot.
        """
        branch_counts_list = [network.branch_counts for network in self.networks]
        labels = [network.name for network in self.networks]
        return parameters.plot_ternary_plot(
            counts_list=branch_counts_list,
            labels=labels,
            is_nodes=False,
            colors=colors,
        )

    def _plot_azimuth_set_lengths(
        self,
        automatic_cut_offs: bool,
        using_branches: bool,
        plot_truncated_data: bool,
    ) -> Tuple[
        Dict[str, length_distributions.MultiLengthDistribution],
        List[length_distributions.Polyfit],
        List[Figure],
        List[Axes],
    ]:
        """
        Plot multi-network azimuths set lengths with fits.
        """
        mlds = self.set_multi_length_distributions(using_branches=using_branches)

        figs, axes, polyfits = [], [], []
        for mld in mlds.values():
            polyfit, fig, ax = mld.plot_multi_length_distributions(
                automatic_cut_offs=automatic_cut_offs,
                plot_truncated_data=plot_truncated_data,
            )
            figs.append(fig)
            axes.append(ax)
            polyfits.append(polyfit)

        return mlds, polyfits, figs, axes

    def plot_trace_azimuth_set_lengths(
        self,
        automatic_cut_offs: bool,
        plot_truncated_data: bool,
    ) -> Tuple[
        Dict[str, length_distributions.MultiLengthDistribution],
        List[length_distributions.Polyfit],
        List[Figure],
        List[Axes],
    ]:
        """
        Plot multi-network trace azimuths set lengths with fits.
        """
        return self._plot_azimuth_set_lengths(
            automatic_cut_offs=automatic_cut_offs,
            using_branches=False,
            plot_truncated_data=plot_truncated_data,
        )

    def plot_branch_azimuth_set_lengths(
        self,
        automatic_cut_offs: bool,
        plot_truncated_data: bool,
    ) -> Tuple[
        Dict[str, length_distributions.MultiLengthDistribution],
        List[length_distributions.Polyfit],
        List[Figure],
        List[Axes],
    ]:
        """
        Plot multi-network trace azimuths set lengths with fits.
        """
        return self._plot_azimuth_set_lengths(
            automatic_cut_offs=automatic_cut_offs,
            using_branches=True,
            plot_truncated_data=plot_truncated_data,
        )

    def set_multi_length_distributions(
        self, using_branches: bool
    ) -> Dict[str, length_distributions.MultiLengthDistribution]:
        """
        Get set-wise multi-length distributions.
        """
        distributions_dict = self.network_length_distributions(
            using_branches=using_branches, using_azimuth_sets=True
        )

        mlds = dict()
        for set_name in self.networks[0].azimuth_set_names:
            set_lengths: List[length_distributions.LengthDistribution] = []
            for lds in distributions_dict.values():
                set_lengths.append(lds[set_name])

            mld = length_distributions.MultiLengthDistribution(
                distributions=set_lengths,
                using_branches=using_branches,
            )
            mlds[set_name] = mld
        return mlds

    def basic_network_descriptions_df(
        self,
        columns: Dict[str, Tuple[Optional[str], Type]],
    ):
        """
        Create DataFrame useful for basic Network characterization.

        ``columns`` should contain key value pairs where the key is the column
        name in ``numerical_network_description`` dict. Value is a tuple where
        the first member is a new name for the column or alternatively None in
        which case the column name isn't changed. The second member should be
        the type of the column, typically either str, int or float.
        """
        numerical_df = pd.DataFrame(
            [network.numerical_network_description() for network in self.networks]
        )
        # Filter to wanted columns only
        numerical_df = numerical_df[list(columns)]

        assert isinstance(numerical_df, pd.DataFrame)

        for column, (_, column_type) in columns.items():
            # apply_func = (
            #     column_type
            #     # if column_type is not float
            #     # else lambda val: round(column_type(val), decimals)
            # )
            numerical_df[column] = numerical_df[column].apply(column_type)

        # Rename columns
        renames = {
            key: value[0] for key, value in columns.items() if value[0] is not None
        }
        numerical_df = numerical_df.rename(renames, axis="columns")

        assert isinstance(numerical_df, pd.DataFrame)

        # Resolve index name
        index_name = NAME
        if NAME in columns:
            new_name = columns[NAME][0]
            if new_name is not None:
                index_name = new_name
        numerical_df.set_index(index_name, inplace=True)

        numerical_df_transposed = numerical_df.transpose()

        return numerical_df_transposed



"""
Direct utilities of trace validation.
"""
import logging
from typing import List, Optional, Tuple

import geopandas as gpd
import numpy as np
from geopandas.sindex import PyGEOSSTRTreeIndex
from shapely.geometry import LineString, MultiLineString, Point, Polygon
from shapely.ops import split

from fractopo.general import (
    geom_bounds,
    safe_buffer,
    spatial_index_intersection,
    within_bounds,
)

log = logging.getLogger(__name__)


def segment_within_buffer(
    linestring: LineString,
    multilinestring: MultiLineString,
    snap_threshold: float,
    snap_threshold_error_multiplier: float,
    overlap_detection_multiplier: float,
) -> bool:
    """
    Check if segment is within buffer of multilinestring.

    First check if given linestring completely overlaps any part of
    multilinestring and if it does, returns True.

    Next it starts to segmentize the multilinestring to smaller
    linestrings and consequently checks if these segments are completely
    within a buffer made of the given linestring.
    It also checks that the segment size is reasonable.

    TODO: segmentize_linestring is very inefficient.
    """
    if multilinestring.is_empty:
        return False
    # Test for a single segment overlap
    if linestring.overlaps(multilinestring):
        return True

    buffered_linestring = safe_buffer(
        linestring, snap_threshold * snap_threshold_error_multiplier
    )
    assert isinstance(linestring, LineString)
    assert isinstance(buffered_linestring, Polygon)
    assert isinstance(multilinestring, MultiLineString)
    assert buffered_linestring.area > 0
    min_x, min_y, max_x, max_y = geom_bounds(buffered_linestring)

    # Crop MultiLineString near to the buffered_linestring
    if not buffered_linestring.intersects(multilinestring):
        return False
    cropped_mls = buffered_linestring.intersection(multilinestring)

    # Check for cases with no chance of stacking
    if cropped_mls.is_empty or (
        isinstance(cropped_mls, LineString)
        and cropped_mls.length < snap_threshold * overlap_detection_multiplier
    ):
        return False

    assert isinstance(cropped_mls, (MultiLineString, LineString))

    all_segments: List[Tuple[Tuple[float, float], Tuple[float, float]]] = []
    ls: LineString

    # Create list of LineStrings from within the crop
    mls_geoms: List[LineString] = (
        list(cropped_mls.geoms)
        if isinstance(cropped_mls, MultiLineString)
        else [cropped_mls]
    )

    # Iterate over list of LineStrings
    for ls in mls_geoms:
        all_segments.extend(
            segmentize_linestring(ls, snap_threshold * overlap_detection_multiplier)
        )
    for start, end in all_segments:
        if within_bounds(
            x=start[0], y=start[1], min_x=min_x, min_y=min_y, max_x=max_x, max_y=max_y
        ) and within_bounds(
            x=end[0],
            y=end[1],
            min_x=min_x,
            min_y=min_y,
            max_x=max_x,
            max_y=max_y
            # *end, min_x, min_y, max_x, max_y
        ):
            ls = LineString([start, end])
            if ls.length > snap_threshold * overlap_detection_multiplier and ls.within(
                buffered_linestring
            ):
                return True
    return False


def segmentize_linestring(
    linestring: LineString, threshold_length: float
) -> List[Tuple[Tuple[float, float], Tuple[float, float]]]:
    """
    Segmentize LineString to smaller parts.

    Resulting parts are not guaranteed to be mergeable back to the original.
    """
    assert isinstance(linestring, LineString)
    segments: List[Tuple[Tuple[float, float], Tuple[float, float]]] = []
    for dist in np.arange(0.0, linestring.length, threshold_length):
        segments.append(linestring_segment(linestring, dist, threshold_length))

    return segments


def linestring_segment(linestring: LineString, dist: float, threshold_length: float):
    """
    Get LineString segment from dist to dist + threshold_length.
    """
    coord_1 = linestring.interpolate(dist).coords[0]
    coord_2 = linestring.interpolate(dist + threshold_length).coords[0]
    return coord_1, coord_2


def split_to_determine_triangle_errors(
    trace: LineString,
    splitter_trace: LineString,
    snap_threshold: float,
    triangle_error_snap_multiplier: float,
) -> bool:
    """
    Split trace with splitter_trace to determine triangle intersections.
    """
    assert isinstance(trace, LineString)
    assert isinstance(splitter_trace, LineString)
    try:
        segments = split(trace, splitter_trace)
    except (ValueError, TypeError):
        trace_intersection = trace.intersection(splitter_trace)
        log.info(
            "Failed to split trace with splitter_trace.",
            extra=dict(
                trace=trace.wkt,
                splitter_trace=splitter_trace.wkt,
                trace_intersection_wkt=(
                    trace_intersection.wkt
                    if hasattr(trace_intersection, "wkt")
                    else "No wkt attribute."
                ),
            ),
            exc_info=True,
        )

        if isinstance(trace_intersection, Point):
            # Splitting can fail in a few different ways e.g. with
            # TypeError: object of type 'LineString' has no len()
            # shapely/**/collection.py\", line 64, in geos_geometrycollection_from_py
            # In this case a simple solution if there's just a single point
            # intersection between traces == No triangle error
            # NOTE: not an exhaustive test

            # Check that case follows expectation... (no overlap)
            assert not trace.overlaps(splitter_trace)
            log.info(
                "Failed to split but intersection was a single point.",
                extra=dict(
                    trace_intersection_wkt=trace_intersection.wkt,
                    trace=trace.wkt,
                    splitter_trace=splitter_trace.wkt,
                    traces_overlap=trace.overlaps(splitter_trace),
                ),
            )
            return False

        # split not possible, the traces overlap
        return True
    if len(segments.geoms) > 2:
        if len(segments.geoms) > 3:
            return True
        middle = determine_middle_in_triangle(
            list(segments.geoms),
            snap_threshold=snap_threshold,
            snap_threshold_error_multiplier=triangle_error_snap_multiplier,
        )
        if len(middle) > 0:
            seg_lengths: List[float] = [seg.length for seg in middle]
        else:
            seg_lengths = [seg.length for seg in segments.geoms]
        for seg_length in seg_lengths:
            if (
                snap_threshold / triangle_error_snap_multiplier
                < seg_length
                < snap_threshold * triangle_error_snap_multiplier
            ):
                return True
    return False


def determine_middle_in_triangle(
    segments: List[LineString],
    snap_threshold: float,
    snap_threshold_error_multiplier: float,
) -> List[LineString]:
    """
    Determine the middle segment within a triangle error.

    The middle segment always intersects the other two.
    """
    candidates = []
    for idx, linestring in enumerate(segments):
        others = segments.copy()
        others.pop(idx)
        if (
            sum(
                linestring.distance(other)
                < snap_threshold * snap_threshold_error_multiplier
                for other in others
            )
            >= 2
        ):
            candidates.append(segments[idx])
    return candidates


def determine_trace_candidates(
    geom: LineString,
    idx: int,
    traces: gpd.GeoDataFrame,
    spatial_index: Optional[PyGEOSSTRTreeIndex],
) -> gpd.GeoSeries:
    """
    Determine potentially intersecting traces with spatial index.
    """
    if spatial_index is None:
        log.error("Expected spatial_index not be None.")
        return gpd.GeoSeries()
    assert isinstance(traces, (gpd.GeoSeries, gpd.GeoDataFrame))
    assert isinstance(spatial_index, PyGEOSSTRTreeIndex)
    candidate_idxs = spatial_index_intersection(spatial_index, geom_bounds(geom))
    candidate_idxs.remove(idx)
    candidate_traces: gpd.GeoSeries = traces.geometry.iloc[candidate_idxs]
    candidate_traces = candidate_traces.loc[  # type: ignore
        [isinstance(geom, LineString) for geom in candidate_traces.geometry.values]
    ]
    return candidate_traces


def is_underlapping(
    geom: LineString,
    trace: LineString,
    endpoint: Point,
    snap_threshold: float,
    snap_threshold_error_multiplier: float,
) -> Optional[bool]:
    """
    Determine if a geom is underlapping.
    """
    split_results = list(split(geom, trace).geoms)
    if len(split_results) == 1:
        # Do not intersect
        return True
    if len(split_results) > 1:
        for segment in split_results:
            if (
                segment.distance(endpoint)
                < snap_threshold * snap_threshold_error_multiplier
            ):
                # Dangling end, overlapping
                return False
    log_prints = {
        "geom": geom,
        "trace": trace,
        "endpoint": endpoint,
        "snap_threshold": snap_threshold,
        "snap_threshold_error_multiplier": snap_threshold_error_multiplier,
    }
    log.error(f"Expected is_underlapping to be resolvable.\nvalues:{log_prints}")
    return None



"""
Utilities for Network subsampling.
"""
import logging
import platform
import random
from itertools import compress, groupby
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np
from joblib import Parallel, delayed
from numpy.random import seed

from fractopo import general
from fractopo.analysis.network import Network
from fractopo.analysis.random_sampling import NetworkRandomSampler, RandomChoice

log = logging.getLogger(__name__)


def create_sample(
    sampler: NetworkRandomSampler,
) -> Optional[Dict[str, Union[general.Number, str]]]:
    """
    Sample with ``NetworkRandomSampler`` and return ``Network`` description.
    """
    # Set random seed instead of inheriting the seed from parent process
    # as is default for multiprocessing
    seed()
    random_sample = sampler.random_network_sample()
    if random_sample.network_maybe is None:
        log.error(
            f"Failed to subsample with sampler {sampler.name}",
        )
        return None
    try:
        result = random_sample.network_maybe.numerical_network_description()
    except Exception:
        # TODO: Add Network info?
        log.error(
            f"Failed to get numerical_network_description from sampler{sampler.name}",
            exc_info=True,
        )
        result = None
    return result


def subsample_networks(
    networks: Sequence[Network],
    min_radii: Union[float, Dict[str, float]],
    random_choice: RandomChoice = RandomChoice.radius,
    samples: int = 1,
) -> List[Optional[Dict[str, Union[general.Number, str]]]]:
    """
    Subsample given Sequence of Networks.
    """
    assert isinstance(samples, int)
    assert samples > 0

    subsamplers = [
        NetworkRandomSampler.random_network_sampler(
            network=network,
            min_radius=min_radii
            if isinstance(min_radii, float)
            else min_radii[network.name],
            random_choice=random_choice,
        )
        for network in networks
    ] * samples

    # Gather subsamples with parallel processing
    subsamples = Parallel(n_jobs=(-1 if not platform.system() == "Windows" else 1))(
        delayed(create_sample)(sampler=sampler) for sampler in subsamplers
    )
    assert isinstance(subsamples, list)

    return subsamples


def gather_subsample_descriptions(
    subsample_results: List[Optional[Dict[str, Union[general.Number, str]]]],
) -> List[Dict[str, Union[general.Number, str]]]:
    """
    Gather results from a list of subsampling ProcessResults.
    """
    descriptions = []
    n_results = len(subsample_results)
    for random_sample_description in subsample_results:
        if random_sample_description is None:
            log.warning("Discarding None subsample.")
            continue
        if not isinstance(random_sample_description, dict):
            log.error(
                "Expected result random_sample_description to be a dict.",
                extra=dict(
                    random_sample_description_type=type(random_sample_description),
                    random_sample_description=random_sample_description,
                ),
            )
            continue

        descriptions.append(random_sample_description)
    n_actual_results = len(descriptions)
    logging.info(f"Out of {n_results}, {n_actual_results} were succesfull subsamples.")
    return descriptions


# def filter_samples_by_area(
#     group: pd.DataFrame, min_area: float, max_area: Optional[float], indexes: List[int]
# ):
#     """
#     Filter samples by given min and max area values.
#     """
#     # Get circle areas
#     areas = group["area"].to_numpy()

#     # Solve max_area
#     max_area = np.max(areas) if max_area is None else max_area

#     # Filter out areas that do not fit within the range
#     area_compressor = [min_area <= area <= max_area for area in areas]

#     # Filter out indexes accordingly
#     indexes = list(compress(indexes, area_compressor))

#     return indexes


def choose_sample_from_group(
    group: general.ParameterListType,
) -> general.ParameterValuesType:
    """
    Choose single sample from group DataFrame.
    """
    # Make continuous index from 0
    indexes = list(range(len(group)))

    assert len(indexes) > 0
    # Choose from indexes
    choice = random.choices(population=indexes, k=1)[0]

    # Get the dict at choice index
    chosen_dict = group[choice]
    assert isinstance(chosen_dict, dict)

    return chosen_dict


def area_weighted_index_choice(
    idxs: List[int], areas: List[float], compressor: List[bool]
) -> int:
    """
    Make area-weighted choce from list of indexes.
    """
    possible_idxs = list(compress(idxs, compressor))
    possible_areas = list(compress(areas, compressor))
    choice = random.choices(population=possible_idxs, weights=possible_areas, k=1)[0]
    return choice


def collect_indexes_of_base_circles(
    idxs: List[int], how_many: int, areas: List[float]
) -> List[int]:
    """
    Collect indexes of base circles, area-weighted and randomly.
    """
    which_idxs = []
    for _ in range(how_many):
        compressor = [idx not in which_idxs for idx in idxs]
        choice = area_weighted_index_choice(
            idxs=idxs, areas=areas, compressor=compressor
        )
        which_idxs.append(choice)

    assert len(which_idxs) == how_many

    return which_idxs


def random_sample_of_circles(
    grouped: Dict[str, general.ParameterListType],
    circle_names_with_diameter: Dict[str, int],
    min_circles: int = 1,
    max_circles: Optional[int] = None,
) -> general.ParameterListType:
    """
    Get a random sample of circles from grouped subsampled data.

    Both the amount of overall circles and which circles within each group
    is random. Data is grouped by target area name.
    """
    if max_circles is not None:
        assert max_circles >= min_circles

    # Area names
    # grouped_keys = grouped.groups.keys()
    # names = [name for name in grouped_keys if isinstance(name, str)]
    # assert len(grouped_keys) == len(names)
    names = list(grouped)

    assert all(name in circle_names_with_diameter for name in names)

    # Get area of the base circles corresponding to area name
    areas = [np.pi * (circle_names_with_diameter[name] / 2) ** 2 for name in names]

    # All indexes
    idxs = list(range(0, len(grouped)))

    # "Randomly" choose how many circles
    # Is constrained by given min_circles and max_circles
    how_many = random.randint(
        min_circles, len(grouped) if max_circles is None else max_circles
    )

    # Collect indexes of base circles
    which_idxs = collect_indexes_of_base_circles(
        idxs=idxs, how_many=how_many, areas=areas
    )

    # Collect the Series that are chosen
    chosen: general.ParameterListType = []

    # Iterate over the DataFrameGroupBy dataframe groups
    for idx, group in enumerate(grouped.values()):
        # Skip if not chosen base circle previously
        if idx not in which_idxs:
            continue

        chosen_dict = choose_sample_from_group(group=group)
        chosen.append(chosen_dict)

    assert len(chosen) == how_many

    # Return chosen subsampled circles from base circles
    return chosen


def aggregate_chosen(
    chosen: general.ParameterListType,
    default_aggregator: general.Aggregator = general.Aggregator.MEAN,
) -> Dict[str, Any]:
    """
    Aggregate a collection of subsampled circles for params.

    Weights averages by the area of each subsampled circle.
    """
    columns = list(chosen[0].keys())

    area_values = [params[general.Param.AREA.value.name] for params in chosen]
    aggregated_values = dict()
    for column in columns:
        aggregator = default_aggregator
        assert callable(aggregator)
        column_values = [params[column] for params in chosen]
        for param in general.Param:
            if column == param.value.name:
                aggregator = param.value.aggregator
                # assert isinstance(aggregator, Callable)
                assert callable(aggregator)
                break
        try:
            aggregated = aggregator(values=column_values, weights=area_values)
        except Exception:
            log.info(
                "Could not aggregate column. Falling back to fallback_aggregation.",
                extra=dict(
                    current_column=column, aggregator=aggregator, columns=columns
                ),
                exc_info=True,
            )
            aggregated = general.fallback_aggregation(values=column_values)
        assert isinstance(aggregated, (int, float, str))
        aggregated_values[column] = aggregated

    return aggregated_values


def groupby_keyfunc(
    item: Dict[str, Union[general.Number, str]],
    groupby_column: str = general.NAME,
) -> str:
    """
    Use groupby_column to group values.
    """
    val = item[groupby_column]
    assert isinstance(val, str)
    return val


def group_gathered_subsamples(
    subsamples: List[Dict[str, Union[general.Number, str]]],
    groupby_column: str = general.NAME,
) -> Dict[str, List[Dict[str, Union[general.Number, str]]]]:
    """
    Group gathered subsamples.

    By default groups by Name column.

    >>> subsamples = [
    ...     {"param": 2.0, "Name": "myname"},
    ...     {"param": 2.0, "Name": "myname"},
    ... ]
    >>> group_gathered_subsamples(subsamples)
    {'myname': [{'param': 2.0, 'Name': 'myname'}, {'param': 2.0, 'Name': 'myname'}]}

    """
    grouped = groupby(
        subsamples,
        key=lambda item: groupby_keyfunc(item=item, groupby_column=groupby_column),
    )
    # for k, g in grouped:
    #     print(k, list(g))

    grouped = {key: list(vals) for key, vals in grouped if isinstance(key, str)}

    return grouped



"""
Trace and branch data analysis with LineData class abstraction.
"""
import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple

import geopandas as gpd
import numpy as np
import pandas as pd
import powerlaw
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.projections import PolarAxes

from fractopo.analysis import azimuth, length_distributions, parameters
from fractopo.general import (
    BOUNDARY_INTERSECT_KEYS,
    Col,
    SetRangeTuple,
    determine_azimuth,
    determine_set,
    intersection_count_to_boundary_weight,
    numpy_to_python_type,
    raise_determination_error,
)

log = logging.getLogger(__name__)


def _column_array_property(
    column: Col,
    gdf: gpd.GeoDataFrame,
) -> Optional[np.ndarray]:
    """
    Return column values from GeoDataFrame if it exists in the GeoDataFrame.
    """
    if column.value in gdf:
        values = gdf[column.value]
        assert isinstance(values, pd.Series)
        result = values.to_numpy()
        assert isinstance(result, np.ndarray)
        return result
    return None


@dataclass
class LineData:

    """
    Wrapper around the given GeoDataFrame with trace or branch data.

    The line_gdf reference is passed and LineData will modify the input
    line_gdf instead of copying the input frame. This means line_gdf columns
    are accessible in the passed input reference upstream.
    """

    _line_gdf: gpd.GeoDataFrame
    using_branches: bool

    azimuth_set_ranges: SetRangeTuple
    azimuth_set_names: Tuple[str, ...]

    length_set_ranges: SetRangeTuple = ()
    length_set_names: Tuple[str, ...] = ()

    area_boundary_intersects: np.ndarray = field(default_factory=lambda: np.array([]))

    _automatic_fit: Optional[powerlaw.Fit] = None

    def __getattr__(self, __name: str) -> Any:
        """
        Overwrite __getattr__ to warn about accessing _line_gdf.
        """
        if __name == "_line_gdf":
            log.error(
                "Output line_gdf might not have all column attributes defined.\n"
                "Use LineData attributes instead of getting the GeoDataFrame."
            )
        return getattr(self, __name)

    def __post_init__(self):
        """
        Initialize LengthDistribution from given data.
        """

    @property
    def azimuth_array(self) -> np.ndarray:
        """
        Array of trace or branch azimuths.
        """
        column_array = _column_array_property(column=Col.AZIMUTH, gdf=self._line_gdf)
        if column_array is None:
            column_array = np.array(
                [determine_azimuth(line, halved=True) for line in self.geometry]
            )
            self._line_gdf[Col.AZIMUTH.value] = column_array
        assert isinstance(column_array, np.ndarray)
        return column_array

    @property
    def azimuth_set_array(self) -> np.ndarray:
        """
        Array of trace or branch azimuth set ids.
        """
        column_array = _column_array_property(
            column=Col.AZIMUTH_SET, gdf=self._line_gdf
        )
        if column_array is None:
            column_array = np.array(
                [
                    determine_set(
                        azimuth,
                        self.azimuth_set_ranges,
                        self.azimuth_set_names,
                        loop_around=True,
                    )
                    for azimuth in self.azimuth_array
                ]
            )
            self._line_gdf[Col.AZIMUTH_SET.value] = column_array
        assert isinstance(column_array, np.ndarray)
        return column_array

    @property
    def length_boundary_weights(self):
        """
        Array of weights for lines based on intersection count with boundary.
        """
        column_array = _column_array_property(
            column=Col.LENGTH_WEIGHTS, gdf=self._line_gdf
        )
        if column_array is None:
            assert self.area_boundary_intersects.dtype in (
                "int64",
                "float64",
                "int32",
                "float32",
            )
            column_array = np.array(
                [
                    intersection_count_to_boundary_weight(int(inter_count))
                    for inter_count in self.area_boundary_intersects
                ]
            )
            self._line_gdf[Col.LENGTH_WEIGHTS.value] = column_array
        return column_array

    @property
    def length_array_non_weighted(self) -> np.ndarray:
        """
        Array of trace or branch lengths not weighted by boundary conditions.
        """
        column_array = _column_array_property(
            column=Col.LENGTH_NON_WEIGHTED, gdf=self._line_gdf
        )
        if column_array is None:
            column_array = self.geometry.length.to_numpy()
            self._line_gdf[Col.LENGTH_NON_WEIGHTED.value] = column_array
        assert isinstance(column_array, np.ndarray)
        return column_array

    @property
    def length_array(self) -> np.ndarray:
        """
        Array of trace or branch lengths.

        Note: lengths can be 0.0 due to boundary weighting.
        """
        column_array = _column_array_property(column=Col.LENGTH, gdf=self._line_gdf)
        if column_array is None:
            new_column_array = (
                self.geometry.length.to_numpy() * self.length_boundary_weights
                if len(self.area_boundary_intersects) > 0
                else 1.0
            )
            self._line_gdf[Col.LENGTH.value] = new_column_array

        else:
            new_column_array = column_array
        assert isinstance(new_column_array, np.ndarray)
        return new_column_array

    @property
    def length_set_array(self) -> np.ndarray:
        """
        Array of trace or branch length set ids.
        """
        if len(self.length_set_names) == 0 or len(self.length_set_ranges) == 0:
            log.error("Expected length_set_names and _ranges to be non-empty.")
            raise_determination_error(
                "length_set_array",
                determine_target="length set attributes",
                verb="initializing",
            )
        column_array = _column_array_property(column=Col.LENGTH_SET, gdf=self._line_gdf)
        if column_array is None:
            column_array = np.array(
                [
                    determine_set(
                        length,
                        self.length_set_ranges,
                        self.length_set_names,
                        loop_around=False,
                    )
                    for length in self.length_array
                ]
            )
            self._line_gdf[Col.LENGTH_SET.value] = column_array
        return column_array

    @property
    def azimuth_set_counts(self) -> Dict[str, int]:
        """
        Get dictionary of azimuth set counts.
        """
        return parameters.determine_set_counts(
            self.azimuth_set_names, self.azimuth_set_array
        )

    @property
    def length_set_counts(self) -> Dict[str, int]:
        """
        Get dictionary of length set counts.
        """
        return parameters.determine_set_counts(
            self.length_set_names, self.length_set_array
        )

    @property
    def automatic_fit(self) -> Optional[powerlaw.Fit]:
        """
        Get automatic powerlaw Fit.
        """
        if self._automatic_fit is None:
            self._automatic_fit = length_distributions.determine_fit(self.length_array)
        return self._automatic_fit

    @property
    def boundary_intersect_count(self) -> Dict[str, int]:
        """
        Get counts of line intersects with boundary.
        """
        assert len(self.area_boundary_intersects) > 0
        keys, counts = np.unique(self.area_boundary_intersects, return_counts=True)
        keys = list(map(str, keys))
        counts = list(map(numpy_to_python_type, counts))
        key_counts = dict(zip(keys, counts))
        for default_key in BOUNDARY_INTERSECT_KEYS:
            if default_key not in key_counts:
                key_counts[default_key] = 0
        assert len(key_counts) == 3
        assert all(np.isin(BOUNDARY_INTERSECT_KEYS, list(key_counts)))
        return key_counts

    @property
    def geometry(self) -> gpd.GeoSeries:
        """
        Get line geometries.
        """
        return self._line_gdf.geometry

    def determine_manual_fit(self, cut_off: float) -> Optional[powerlaw.Fit]:
        """
        Get manually determined Fit with set cut off.
        """
        return length_distributions.determine_fit(self.length_array, cut_off=cut_off)

    def describe_fit(
        self, label: Optional[str] = None, cut_off: Optional[float] = None
    ):
        """
        Return short description of automatic powerlaw fit.
        """
        fit = (
            self.automatic_fit
            if cut_off is None
            else self.determine_manual_fit(cut_off=cut_off)
        )
        return length_distributions.describe_powerlaw_fit(
            fit, label=label, length_array=self.length_array
        )

    def plot_lengths(
        self,
        label: str,
        use_probability_density_function: bool,
        fit: Optional[powerlaw.Fit] = None,
        plain: bool = False,
    ) -> Tuple[Optional[powerlaw.Fit], Figure, Axes]:
        """
        Plot length data with powerlaw fit.
        """
        return length_distributions.plot_distribution_fits(
            self.length_array,
            label=label,
            fit=self.automatic_fit if fit is None else fit,
            using_branches=self.using_branches,
            use_probability_density_function=use_probability_density_function,
            plain=plain,
        )

    def plot_azimuth(
        self,
        label: str,
        append_azimuth_set_text: bool = False,
        add_abundance_order: bool = False,
        visualize_sets: bool = False,
        bar_color: str = "darkgrey",
        plain: bool = False,
    ) -> Tuple[azimuth.AzimuthBins, Figure, PolarAxes]:
        """
        Plot azimuth data in rose plot.
        """
        return azimuth.plot_azimuth_plot(
            self.azimuth_array,
            self.length_array,
            self.azimuth_set_array,
            self.azimuth_set_names,
            azimuth_set_ranges=self.azimuth_set_ranges,
            label=label,
            append_azimuth_set_text=append_azimuth_set_text,
            visualize_sets=visualize_sets,
            bar_color=bar_color,
            plain=plain,
            add_abundance_order=add_abundance_order,
        )

    def plot_azimuth_set_count(self, label: str) -> Tuple[Figure, Axes]:
        """
        Plot azimuth set counts.
        """
        return parameters.plot_set_count(self.azimuth_set_counts, label=label)

    def plot_length_set_count(self, label: str) -> Tuple[Figure, Axes]:
        """
        Plot length set counts.
        """
        return parameters.plot_set_count(self.length_set_counts, label=label)

    def boundary_intersect_count_desc(self, label: str) -> Dict[str, int]:
        """
        Get counts of line intersects with boundary.
        """
        key_counts = self.boundary_intersect_count
        intersect_key_counts = dict()
        for key, item in key_counts.items():
            # trace_key_counts[f"Trace Boundary {key} Intersect Count"] = item
            intersect_key_counts[f"{label} Boundary {key} Intersect Count"] = item
        return intersect_key_counts

    @property
    def azimuth_set_length_arrays(self) -> Dict[str, np.ndarray]:
        """
        Get length arrays of each azimuth set.
        """
        return {
            azimuth_set_name: self.length_array[
                self.azimuth_set_array == azimuth_set_name
            ]
            for azimuth_set_name in self.azimuth_set_names
        }

    def plot_azimuth_set_lengths(
        self,
        use_probability_density_function: bool = False,
    ) -> Tuple[List[Optional[powerlaw.Fit]], List[Figure], List[Axes]]:
        """
        Plot azimuth set length distributions with fits.
        """
        # Collect fits and plots
        fits, figs, axes = [], [], []

        # Iterate over azimuth set names and corresponding length arrays
        for azimuth_set_name, set_lengths in self.azimuth_set_length_arrays.items():
            # Create plot for each set length array
            fit, fig, ax = length_distributions.plot_distribution_fits(
                set_lengths,
                label=azimuth_set_name,
                using_branches=self.using_branches,
                use_probability_density_function=use_probability_density_function,
            )

            # Collect
            fits.append(fit)
            figs.append(fig)
            axes.append(ax)

        return fits, figs, axes
